import os
import re
import tempfile
import time

from Ganga.GPIDev.Credentials import getCredential

from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.logging import getLogger

from Ganga.Utility.GridShell import getShell

from Ganga.Lib.LCG.GridftpSandboxCache import GridftpFileIndex, GridftpSandboxCache

from Ganga.Lib.LCG.Utility import get_uuid
from Ganga.Lib.Root import randomString

# global variables
logger = getLogger()


class Grid(object):

    '''Helper class to implement grid interaction'''

    # attributes
    _attributes = ('middleware', 'credential', 'config', 'active', 'perusable')

    def __init__(self, middleware='EDG'):

        self.active = False

        self.re_token = re.compile('^token:(.*):(.*)$')

        self.credential = None

        self.middleware = middleware.upper()

        self.perusable = False

        self.config = getConfig('LCG')

        self.wms_list = []

        self.new_config = ""

#       check that UI has been set up
#       start up a shell object specific to the middleware
        self.shell = getShell(self.middleware)

        self.proxy_id = {}

        if not self.shell:
            logger.warning(
                'LCG-%s UI has not been configured. The plugin has been disabled.' % self.middleware)
            return

#       create credential for this Grid object
        self.active = self.check_proxy()

    def __setattr__(self, attr, value):
        object.__setattr__(self, attr, value)
        # dynamic update the internal shell object if the config attribute is
        # reset
        if attr == 'config':
            self.shell = getShell(self.middleware)

    def __get_cmd_prefix_hack__(self, binary=False):
        # this is to work around inconsistency of LCG setup script and commands:
        # LCG commands require python2.2 but the setup script does not set this version of python
        # if another version of python is used (like in GUI), then python2.2 runs against wrong python libraries
        # possibly should be fixed in LCG: either remove python2.2 from command scripts or make setup script force
        # correct version of python
        prefix_hack = "${%s_LOCATION}/bin/" % self.middleware

        # some script-based glite-wms commands (status and logging-info) requires (#/usr/bin/env python2)
        # which leads to a python conflict problem.
        if not binary:
            prefix_hack = 'python ' + prefix_hack

        return prefix_hack

    def __set_submit_option__(self):

        #       find out how the VO has been specified

        submit_option = ''

        msg = 'use VO defined '

        voms = self.__get_proxy_voname__()

        # VO specific WMS options (no longer used by glite-wms-job-submit command)
        # 1. vo specified in the configuration file
        if self.config['ConfigVO']:
            if self.middleware == 'EDG':
                submit_option = '--config-vo %s' % self.config['ConfigVO']
                if not os.path.exists(self.config['ConfigVO']):
                    #raise Ganga.Utility.Config.ConfigError('')
                    raise ConfigError('')
                else:
                    msg += 'in %s.' % self.config['ConfigVO']
            else:
                logger.warning(
                    'ConfigVO configuration ignored by %s middleware. Set Config instead.' % self.middleware)
        # 2. vo attached in the voms proxy
        elif voms:
            msg += 'in voms proxy: %s.' % voms
        # 3. vo is given explicitely
        elif self.config['VirtualOrganisation']:
            submit_option = '--vo %s' % self.config['VirtualOrganisation']
            msg += 'in Ganga config: %s.' % self.config['VirtualOrganisation']
        # 4. no vo information is found
        else:
            logger.warning(
                'No Virtual Organisation specified in the configuration. The plugin has been disabeled.')
            return None

        # check for WMS list - has it changed?
        if self.config['GLITE_ALLOWED_WMS_LIST'] == []:
            self.wms_list = self.config['GLITE_ALLOWED_WMS_LIST'][:]

            # remove old conf
            if self.new_config != "" and os.path.exists(self.new_config):
                try:
                    os.remove(self.new_config)
                except:
                    logger.warning(
                        "Unable to remove old config file '%s'" % self.new_config)

            self.new_config = ""
        elif self.wms_list != self.config['GLITE_ALLOWED_WMS_LIST']:
            logger.info("Updating allowed WMS list...")

            # find path to wms conf file and a temporary file to copy to
            wms_conf_path = os.path.join(self.shell.env['GLITE_WMS_LOCATION'], 'etc', self.config[
                                         'VirtualOrganisation'], 'glite_wmsui.conf')
            temp_wms_conf = tempfile.mktemp('.conf')

            # read in original
            with open(wms_conf_path, "r") as this_file:
                orig_text = this_file.read()

            # find the last bracket and add in the new text
            pos = orig_text.rfind("]")
            wms_text = "\nWMProxyEndpoints  =  {" + ",".join(
                ["\"%s\"" % wms for wms in self.config['GLITE_ALLOWED_WMS_LIST']]) + "};\n]\n"
            new_text = orig_text[:pos] + wms_text

            # write the new config file
            with open(temp_wms_conf, "w") as this_file:
                this_file.write(new_text)

            self.wms_list = self.config['GLITE_ALLOWED_WMS_LIST'][:]

            # remove old conf
            if self.new_config != "" and os.path.exists(self.new_config):
                try:
                    os.remove(self.new_config)
                except:
                    logger.warning(
                        "Unable to remove old config file '%s'" % self.new_config)

            self.new_config = temp_wms_conf

        # general WMS options
        # NB. please be aware the config for gLite WMS is NOT compatible with the config for EDG RB
        #     although both shares the same command option: '--config'
        if self.config['Config']:
            submit_option += ' --config %s' % self.config['Config']
        elif self.new_config:
            submit_option += ' --config %s' % self.new_config

        submit_option = ' %s ' % submit_option

        logger.debug(msg)

        return submit_option

    def __resolve_gridcmd_log_path__(self, regxp_logfname, cmd_output):
        match_log = re.search(regxp_logfname, cmd_output)

        logfile = None
        if match_log:
            logfile = match_log.group(1)
        return logfile

    def __clean_gridcmd_log__(self, regxp_logfname, cmd_output):

        logfile = self.__resolve_gridcmd_log_path__(regxp_logfname, cmd_output)

        if logfile and os.path.exists(logfile):
            os.remove(logfile)

        return True

    def __print_gridcmd_log__(self, regxp_logfname, cmd_output):

        logfile = self.__resolve_gridcmd_log_path__(regxp_logfname, cmd_output)

        if logfile:
            for l in open(logfile, 'r'):
                logger.warning(l.strip())

            # here we assume the logfile is no longer needed at this point -
            # remove it
            os.remove(logfile)
        else:
            logger.warning('output\n%s\n', cmd_output)
            logger.warning('end of output')

    def __get_proxy_voname__(self):
        '''Check validity of proxy vo'''

        # for EDG, we never check it
        if self.middleware == 'EDG':
            return None
        else:
            logger.debug('voms of credential: %s' % self.credential.voms)
            return self.credential.voms

    def __get_lfc_host__(self):
        '''Gets the LFC_HOST: from current shell or querying BDII on demand'''
        lfc_host = None

        if 'LFC_HOST' in self.shell.env:
            lfc_host = self.shell.env['LFC_HOST']

        if not lfc_host:
            lfc_host = self.__get_default_lfc__()

        return lfc_host

    def __get_default_lfc__(self):
        '''Gets the default lfc host from lcg-infosites'''

        cmd = 'lcg-infosites'

        logger.debug('%s lfc-infosites called ...' % self.middleware)

        rc, output, m = self.shell.cmd1(
            '%s --vo %s lfc' % (cmd, self.config['VirtualOrganisation']), allowed_exit=[0, 255])

        if rc != 0:
            # self.__print_gridcmd_log__('lcg-infosites',output)
            return None
        else:
            lfc_list = output.strip().split('\n')
            return lfc_list[0]

    def __resolve_no_matching_jobs__(self, cmd_output):
        '''Parsing the glite-wms-job-status log to get the glite jobs which have been removed from the WMS'''

        logfile = self.__resolve_gridcmd_log_path__(
            '(.*-job-status.*\.log)', cmd_output)

        glite_ids = []

        if logfile:

            re_jid = re.compile(
                '^Unable to retrieve the status for: (https:\/\/\S+:9000\/[0-9A-Za-z_\.\-]+)\s*$')
            re_key = re.compile('^.*(no matching jobs found)\s*$')

            m_jid = None
            m_key = None
            myjid = ''
            for line in open(logfile, 'r'):
                m_jid = re_jid.match(line)
                if m_jid:
                    myjid = m_jid.group(1)
                    m_jid = None

                if myjid:
                    m_key = re_key.match(line)
                    if m_key:
                        glite_ids.append(myjid)
                        myjid = ''

        return glite_ids

    def check_proxy(self):
        '''Check the proxy and prompt the user to refresh it'''

        if self.credential is None:
            self.credential = getCredential('GridProxy', self.middleware)

        if self.middleware == 'GLITE':
            self.credential.voms = self.config['VirtualOrganisation']
            self.credential = getCredential('GridProxy', 'GLITE')

        status = self.credential.renew(maxTry=3)

        if not status:
            logger.warning("Could not get a proxy, giving up after 3 retries")
            return False

        return True

    def list_match(self, jdlpath, ce=None):
        '''Returns a list of computing elements can run the job'''

        re_ce = re.compile('^\s*\-\s*(\S+\:(2119|8443)\/\S+)\s*$')

        matched_ces = []

        if self.middleware == 'EDG':
            cmd = 'edg-job-list-match'
            exec_bin = False
        else:
            cmd = 'glite-wms-job-list-match -a'
            exec_bin = True

        if not self.active:
            logger.warning('LCG plugin not active.')
            return

        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return

        submit_opt = self.__set_submit_option__()

        if not submit_opt:
            return matched_ces
        else:
            cmd += submit_opt

        cmd = '%s --noint %s' % (cmd, jdlpath)

        logger.debug('job list match command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        for l in output.split('\n'):

            matches = re_ce.match(l)

            if matches:
                matched_ces.append(matches.group(1))

        if ce:
            if matched_ces.count(ce) > 0:
                matched_ces = [ce]
            else:
                matched_ces = []

        logger.debug('== matched CEs ==')
        for myce in matched_ces:
            logger.debug(myce)
        logger.debug('== matched CEs ==')

        return matched_ces

    def submit(self, jdlpath, ce=None):
        '''Submit a JDL file to LCG'''

        # doing job submission
        if self.middleware == 'EDG':
            cmd = 'edg-job-submit'
            exec_bin = False
        else:
            cmd = 'glite-wms-job-submit -a'
            exec_bin = True

        if not self.active:
            logger.warning('LCG plugin not active.')
            return

        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return

        submit_opt = self.__set_submit_option__()

        if not submit_opt:
            return
        else:
            cmd += submit_opt

        if ce:
            cmd = cmd + ' -r %s' % ce

        cmd = '%s --nomsg %s < /dev/null' % (cmd, jdlpath)

        logger.debug('job submit command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['SubmissionTimeout'])

        if output:
            output = "%s" % output.strip()

        #match = re.search('^(https:\S+)',output)
        match = re.search('.*(https:\/\/\S+:9000\/[0-9A-Za-z_\.\-]+)', output)

        if match:
            logger.debug('job id: %s' % match.group(1))
            if self.middleware == 'GLITE' and self.perusable:
                logger.info("Enabling perusal")
                per_rc, per_out, per_m = self.shell.cmd1(
                    "glite-wms-job-perusal --set -f stdout %s" % match.group(1))

            # remove the glite command log if it exists
            self.__clean_gridcmd_log__('(.*-job-submit.*\.log)', output)
            return match.group(1)

        else:
            logger.warning('Job submission failed.')
            self.__print_gridcmd_log__('(.*-job-submit.*\.log)', output)
            return

    def native_master_cancel(self, jobids):
        '''Native bulk cancellation supported by GLITE middleware.'''

        if self.middleware == 'EDG':
            logger.warning(
                'EDG middleware doesn\'t support bulk cancellation.')
            return False
        else:
            cmd = 'glite-wms-job-cancel'
            exec_bin = True

        if not self.active:
            logger.warning('LCG plugin not active.')
            return False

        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        if not self.__set_submit_option__():
            return False

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        cmd = '%s --noint -i %s' % (cmd, idsfile)

        logger.debug('job cancel command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        if rc != 0:
            logger.warning('Job cancellation failed.')
            self.__print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return False
        else:
            # job cancellation succeeded, try to remove the glite command
            # logfile if it exists
            self.__clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return True

    def status(self, jobids, is_collection=False):
        '''Query the status of jobs on the grid'''

        if not jobids:
            return ([], [])

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        if self.middleware == 'EDG':
            cmd = 'edg-job-status'
            exec_bin = False
        else:
            cmd = 'glite-wms-job-status'

            exec_bin = True
            if self.config['IgnoreGliteScriptHeader']:
                exec_bin = False

            if is_collection:
                cmd = '%s -v 3' % cmd

        if not self.active:
            logger.warning('LCG plugin not active.')
            return ([], [])
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return ([], [])

        cmd = '%s --noint -i %s' % (cmd, idsfile)
        logger.debug('job status command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['StatusPollingTimeout'])
        os.remove(idsfile)

        missing_glite_jids = []
        if rc != 0:
            missing_glite_jids = self.__resolve_no_matching_jobs__(output)

            if missing_glite_jids:
                logger.info(
                    'some jobs removed from WMS, will set corresponding Ganga job to \'failed\' status')
                logger.debug('jobs removed from WMS: %s' %
                             repr(missing_glite_jids))
            else:
                self.__print_gridcmd_log__('(.*-job-status.*\.log)', output)

        # job status query succeeded, try to remove the glite command logfile
        # if it exists
        self.__clean_gridcmd_log__('(.*-job-status.*\.log)', output)

        re_id = re.compile('^\s*Status info for the Job : (https://.*\S)\s*$')
        re_status = re.compile('^\s*Current Status:\s+(.*\S)\s*$')

        # from glite UI version 1.5.14, the attribute 'Node Name:' is no longer available
        # for distinguishing master and node jobs. A new way has to be applied.
        #re_name = re.compile('^\s*Node Name:\s+(.*\S)\s*$')
        re_exit = re.compile('^\s*Exit code:\s+(.*\S)\s*$')
        re_reason = re.compile('^\s*Status Reason:\s+(.*\S)\s*$')
        re_dest = re.compile('^\s*Destination:\s+(.*\S)\s*$')

        # pattern to distinguish master and node jobs
        re_master = re.compile('^BOOKKEEPING INFORMATION:\s*$')
        re_node = re.compile('^- Nodes information.*\s*$')

        # pattern for node jobs
        re_nodename = re.compile('^\s*NodeName\s*=\s*"(gsj_[0-9]+)";\s*$')

        info = []
        is_master = False
        is_node = False
        #node_cnt  = 0
        for line in output.split('\n'):

            match = re_master.match(line)
            if match:
                is_master = True
                is_node = False
                #node_cnt  = 0
                continue

            match = re_node.match(line)
            if match:
                is_master = False
                is_node = True
                continue

            match = re_id.match(line)
            if match:
                info += [{'id': match.group(1),
                          'name': '',
                          'is_node': False,
                          'status': '',
                          'exit': '',
                          'reason': '',
                          'destination': ''}]
                if is_node:
                    info[-1]['is_node'] = True
                # if is_node:
                #    info[-1]['name'] = 'node_%d' % node_cnt
                #    node_cnt = node_cnt + 1
                continue

            match = re_nodename.match(line)
            if match and is_node:
                info[-1]['name'] = match.group(1)
                #logger.debug('id: %s, name: %s' % (info[-1]['id'],info[-1]['name']))
                continue

            match = re_status.match(line)
            if match:
                info[-1]['status'] = match.group(1)
                continue

            match = re_exit.match(line)
            if match:
                info[-1]['exit'] = match.group(1)
                continue

            match = re_reason.match(line)
            if match:
                info[-1]['reason'] = match.group(1)
                continue

            match = re_dest.match(line)
            if match:
                info[-1]['destination'] = match.group(1)
                continue

        return (info, missing_glite_jids)

    def get_loginfo(self, jobids, directory, verbosity=1):
        '''Fetch the logging info of the given job and save the output in the job's outputdir'''

        if self.middleware == 'EDG':
            cmd = 'edg-job-get-logging-info -v %d' % verbosity
            exec_bin = False
        else:
            cmd = 'glite-wms-job-logging-info -v %d' % verbosity

            exec_bin = True
            if self.config['IgnoreGliteScriptHeader']:
                exec_bin = False

        if not self.active:
            logger.warning('LCG plugin not active.')
            return False
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        log_output = directory + '/__jobloginfo__.log'

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        cmd = '%s --noint -o %s -i %s' % (cmd, log_output, idsfile)

        logger.debug('job logging info command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])
        os.remove(idsfile)

        if rc != 0:
            self.__print_gridcmd_log__('(.*-logging-info.*\.log)', output)
            return False
        else:
            # logging-info checking succeeded, try to remove the glite command
            # logfile if it exists
            self.__clean_gridcmd_log__('(.*-logging-info.*\.log)', output)
            # returns the path to the saved logging info if success
            return log_output

    def get_output(self, jobid, directory, wms_proxy=False):
        '''Retrieve the output of a job on the grid'''

        if self.middleware == 'EDG':
            cmd = 'edg-job-get-output'
            exec_bin = False
        else:
            cmd = 'glite-wms-job-output'
            exec_bin = True
            # general WMS options (somehow used by the glite-wms-job-output
            # command)
            if self.config['Config']:
                cmd += ' --config %s' % self.config['Config']

        if not self.active:
            logger.warning('LCG plugin is not active.')
            return (False, None)
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return (False, None)

        cmd = '%s --noint --dir %s %s' % (cmd, directory, jobid)

        logger.debug('job get output command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        match = re.search('directory:\n\s*(\S+)\s*\n', output)

        if not match:
            logger.warning('Job output fetch failed.')
            self.__print_gridcmd_log__('(.*-output.*\.log)', output)
            return (False, 'cannot fetch job output')

        # job output fetching succeeded, try to remove the glite command
        # logfile if it exists
        self.__clean_gridcmd_log__('(.*-output.*\.log)', output)

        outdir = match.group(1)

#       some versions of LCG middleware create an extra output directory (named <uid>_<jid_hash>)
#       inside the job.outputdir. Try to match the jid_hash in the outdir. Do output movememnt
#       if the <jid_hash> is found in the path of outdir.
        import urlparse
        jid_hash = urlparse.urlparse(jobid)[2][1:]

        if outdir.count(jid_hash):
            if self.shell.system('mv %s/* %s' % (outdir, directory)) == 0:
                try:
                    os.rmdir(outdir)
                except Exception as msg:
                    logger.warning(
                        "Error trying to remove the empty directory %s:\n%s" % (outdir, msg))
            else:
                logger.warning("Error moving output from %s to %s.\nOutput is left in %s." % (
                    outdir, directory, outdir))
        else:
            pass

        return self.__get_app_exitcode__(directory)

    def cancelMultiple(self, jobids):
        '''Cancel multiple jobs in one LCG job cancellation call'''

        # compose a temporary file with job ids in it
        if not jobids:
            return True

        # do the cancellation using a proper LCG command
        if self.middleware == 'EDG':
            cmd = 'edg-job-cancel'
            exec_bin = False
        else:
            cmd = 'glite-wms-job-cancel'
            exec_bin = True

        if not self.active:
            logger.warning('LCG plugin is not active.')
            return False
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        # compose the cancel command
        cmd = '%s --noint -i %s' % (cmd, idsfile)

        logger.debug('job cancel command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        if rc == 0:
            # job cancelling succeeded, try to remove the glite command logfile
            # if it exists
            self.__clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return True
        else:
            logger.warning("Failed to cancel jobs.\n%s" % output)
            self.__print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return False

    def cancel(self, jobid):
        '''Cancel a job'''

        if self.middleware == 'EDG':
            cmd = 'edg-job-cancel'
            exec_bin = False
        else:
            cmd = 'glite-wms-job-cancel'
            exec_bin = True

        if not self.active:
            logger.warning('LCG plugin is not active.')
            return False
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        cmd = '%s --noint %s' % (cmd, jobid)

        logger.debug('job cancel command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        if rc == 0:
            # job cancelling succeeded, try to remove the glite command logfile
            # if it exists
            self.__clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return True
        else:
            logger.warning("Failed to cancel job %s.\n%s" % (jobid, output))
            self.__print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return False

    def __cream_parse_job_status__(self, log):
        '''Parsing job status report from CREAM CE status query'''

        jobInfoDict = {}

        re_jid = re.compile('^\s+JobID\=\[(https:\/\/.*[:0-9]?\/CREAM.*)\]$')
        re_log = re.compile('^\s+(\S+.*\S+)\s+\=\s+\[(.*)\]$')

        re_jts = re.compile('^\s+Job status changes:$')
        re_ts = re.compile(
            '^\s+Status\s+\=\s+\[(.*)\]\s+\-\s+\[(.*)\]\s+\(([0-9]+)\)$')
        re_cmd = re.compile('^\s+Issued Commands:$')

        # in case of status retrival failed
        re_jnf = re.compile('^.*job not found.*$')

        jid = None

        for jlog in log.split('******')[1:]:

            for l in jlog.split('\n'):
                l.strip()

                m = re_jid.match(l)

                if m:
                    jid = m.group(1)
                    jobInfoDict[jid] = {}
                    continue

                if re_jnf.match(l):
                    break

                m = re_log.match(l)
                if m:
                    att = m.group(1)
                    val = m.group(2)
                    jobInfoDict[jid][att] = val
                    continue

                if re_jts.match(l):
                    jobInfoDict[jid]['Timestamps'] = {}
                    continue

                m = re_ts.match(l)
                if m:
                    s = m.group(1)
                    t = int(m.group(3))
                    jobInfoDict[jid]['Timestamps'][s] = t
                    continue

                if re_cmd.match(l):
                    break

        return jobInfoDict

    def __cream_ui_check__(self):
        '''checking if CREAM CE environment is set properly'''

        if self.middleware.upper() != 'GLITE':
            logger.warning('CREAM CE job operation not supported')
            return False

        if not self.active:
            logger.warning('LCG plugin not active.')
            return False

        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        return True

    def cream_check_delegated_proxy(self, ce):
        '''
        Checking if the delegation id to given CE is still valid.
        '''

        id = None

        try:
            id, t_expire = self.proxy_id[ce]

            # TODO: implement the check of the validity of the delegation id
            now = time.time()

            # check if the lifetime of the existing proxy still longer than
            # 1800 seconds
            if t_expire - now <= 1800:
                logger.debug('existing proxy going to expire in 1800 seconds.')
                id = None
            else:
                logger.debug('reusing valid proxy %s on %s' % (id, ce))

        except KeyError:
            pass

        return id

    def cream_proxy_delegation(self, ce):
        '''CREAM CE proxy delegation'''

        if not self.__cream_ui_check__():
            return

        if not ce:
            logger.warning('No CREAM CE endpoint specified')
            return

        mydelid = self.cream_check_delegated_proxy(ce)

        if not mydelid:

            logger.debug('making new proxy delegation to %s' % ce)

            t_expire = 0

            exec_bin = True

            cmd = 'glite-ce-delegate-proxy'

            cmd += ' -e %s' % ce.split('/cream')[0]

            mydelid = '%s_%s' % (self.credential.identity(), get_uuid())

            cmd = '%s "%s"' % (cmd, mydelid)

            logger.debug('proxy delegation command: %s' % cmd)

            rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                            allowed_exit=[0, 255],
                                            timeout=self.config['SubmissionTimeout'])
            if rc != 0:
                # failed to delegate proxy
                logger.error('proxy delegation error: %s' % output)
                mydelid = ''
            else:
                # proxy delegated successfully
                # NB: expiration time is "current time" + "credential lifetime"
                t_expire = time.time() + \
                    float(
                        self.credential.timeleft(units="seconds", force_check=True))

                logger.debug('new proxy at %s valid until %s' %
                             (ce, time.ctime(t_expire)))

            self.proxy_id[ce] = [mydelid, t_expire]

        return mydelid

    def cream_submit(self, jdlpath, ce):
        '''CREAM CE direct job submission'''

        if not self.__cream_ui_check__():
            return

        if not ce:
            logger.warning('No CREAM CE endpoint specified')
            return

        cmd = 'glite-ce-job-submit'
        exec_bin = True

        mydelid = self.cream_proxy_delegation(ce)

        if mydelid:
            cmd = cmd + ' -D "%s"' % mydelid
        else:
            cmd = cmd + ' -a'

        cmd = cmd + ' -r %s' % ce

        cmd = '%s --nomsg %s < /dev/null' % (cmd, jdlpath)

        logger.debug('job submit command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['SubmissionTimeout'])

        if output:
            output = "%s" % output.strip()

        match = re.search('^(https:\/\/\S+:8443\/[0-9A-Za-z_\.\-]+)$', output)

        if match:
            logger.debug('job id: %s' % match.group(1))
            return match.group(1)
        else:
            logger.warning('Job submission failed.')
            return

    def cream_status(self, jobids):
        '''CREAM CE job status query'''

        if not self.__cream_ui_check__():
            return ([], [])

        if not jobids:
            return ([], [])

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('##CREAMJOBS##\n' + '\n'.join(jobids) + '\n')

        cmd = 'glite-ce-job-status'
        exec_bin = True

        cmd = '%s -L 2 -n -i %s' % (cmd, idsfile)
        logger.debug('job status command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['StatusPollingTimeout'])
        jobInfoDict = {}
        if rc == 0 and output:
            jobInfoDict = self.__cream_parse_job_status__(output)

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        return jobInfoDict

    def cream_purgeMultiple(self, jobids):
        '''CREAM CE job purging'''

        if not self.__cream_ui_check__():
            return False

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('##CREAMJOBS##\n' + '\n'.join(jobids) + '\n')

        cmd = 'glite-ce-job-purge'
        exec_bin = True

        cmd = '%s -n -N -i %s' % (cmd, idsfile)

        logger.debug('job purge command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        logger.debug(output)

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        if rc == 0:
            return True
        else:
            return False

    def cream_cancelMultiple(self, jobids):
        '''CREAM CE job cancelling'''

        if not self.__cream_ui_check__():
            return False

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('##CREAMJOBS##\n' + '\n'.join(jobids) + '\n')

        cmd = 'glite-ce-job-cancel'
        exec_bin = True

        cmd = '%s -n -N -i %s' % (cmd, idsfile)

        logger.debug('job cancel command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        logger.debug(output)

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        if rc == 0:
            return True
        else:
            return False

    def cream_get_output(self, osbURIList, directory):
        '''CREAM CE job output retrieval'''

        if not self.__cream_ui_check__():
            return (False, None)

        gfiles = []
        for uri in osbURIList:
            gf = GridftpFileIndex()
            gf.id = uri
            gfiles.append(gf)

        cache = GridftpSandboxCache()
        cache.middleware = 'GLITE'
        cache.vo = self.config['VirtualOrganisation']
        cache.uploaded_files = gfiles

        return cache.download(files=map(lambda x: x.id, gfiles), dest_dir=directory)

    def __get_app_exitcode__(self, outputdir):

        import Ganga.Core.Sandbox as Sandbox

        Sandbox.getPackedOutputSandbox(outputdir, outputdir)

        # check the application exit code
        app_exitcode = -1
        runtime_log = os.path.join(outputdir, '__jobscript__.log')
        pat = re.compile(r'.*exit code (\d+).')

        if not os.path.exists(runtime_log):
            logger.warning('job runtime log not found: %s' % runtime_log)
            return (False, 'job runtime log not found: %s' % runtime_log)

        for line in open(runtime_log, 'r'):
            mat = pat.match(line)
            if mat:
                app_exitcode = eval(mat.groups()[0])
                break

        # returns False if the exit code of the real executable is not zero
        # the job status of GANGA will be changed to 'failed' if the return
        # value is False
        if app_exitcode != 0:
            logger.debug(
                'job\'s executable returns non-zero exit code: %d' % app_exitcode)
            return (False, app_exitcode)
        else:
            return (True, 0)

    @staticmethod
    def expandxrsl(items):
        '''Expand xrsl items'''

        xrsl = "&\n"
        for key, value in items.iteritems():

            if key == "inputFiles":
                # special case for input files
                xrsl += "(inputFiles="

                for f in value:
                    xrsl += "(\"%s\" \"%s\")\n" % (os.path.basename(f), f)

                xrsl += ")\n"

            elif key == "outputFiles":
                # special case for input files
                xrsl += "(outputFiles="

                for f in value:
                    xrsl += "(\"%s\" \"\")\n" % (os.path.basename(f))

                xrsl += ")\n"

            elif isinstance(value, dict):
                # expand if a dictionary
                xrsl += "(%s=" % key
                for key2, value2 in value.iteritems():
                    xrsl += "(\"%s\" \"%s\")\n" % (key2, value2)

                xrsl += ")\n"
            else:
                # straight key pair
                xrsl += "(%s=\"%s\")\n" % (key, value)

        return xrsl

    @staticmethod
    def expandjdl(items):
        '''Expand jdl items'''

        text = "[\n"
        for key, value in items.iteritems():

            if key == 'Requirements':
                if value:
                    text += 'Requirements = \n   %s;\n' % ' &&\n   '.join(
                        value)

            elif key in ['ShallowRetryCount', 'RetryCount', 'NodeNumber', 'ExpiryTime', 'PerusalTimeInterval']:
                try:
                    value = int(value)
                    if value < 0:
                        raise ValueError
                    text += '%s = %d;\n' % (key, value)
                except ValueError:
                    logger.warning('%s is not positive integer.' % key)

            elif key == 'Environment':
                if value:
                    text += 'Environment = {\n   "%s"\n};\n' % '",\n   "'.join(
                        ['%s=\'%s\'' % var for var in value.items()])

            elif key == 'DataRequirements':
                text += 'DataRequirements = {\n'
                for entry in value:
                    text += '  [\n'
                    text += '    InputData = {\n'
                    for datafile in entry['InputData']:
                        text += '      "%s",\n' % datafile
                    # Get rid of trailing comma
                    text = text.rstrip(',\n') + '\n'
                    text += '    };\n'
                    text += '    DataCatalogType = "%s";\n' % entry[
                        'DataCatalogType']
                    if 'DataCatalog' in entry:
                        text += '    DataCatalog = "%s";\n' % entry[
                            'DataCatalog']
                    text += '  ],\n'
                text = text.rstrip(',\n') + '\n'  # Get rid of trailing comma
                text += '};\n'

            elif isinstance(value, list):
                if value:
                    text += '%s = {\n   "%s"\n};\n' % (key,
                                                       '",\n   "'.join(value))

            elif key == 'Rank':
                text += 'Rank = ( %s );\n' % value

            elif key == 'Nodes':
                text += 'Nodes = %s;\n' % value

            elif key in ['PerusalFileEnable', 'AllowZippedISB']:
                text += '%s = %s;\n' % (key, value)

            else:
                text += '%s = "%s";\n' % (key, value)

        text += "\n]\n"
        return text

    def wrap_lcg_infosites(self, opts=""):
        '''Wrap the lcg-infosites command'''

        cmd = 'lcg-infosites --vo %s %s' % (
            self.config['VirtualOrganisation'], opts)

        if not self.active:
            logger.warning('LCG plugin not active.')
            return

        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return

        logger.debug('lcg-infosites command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s' % (cmd), allowed_exit=[0, 255])

        if rc != 0:
            return ""
        else:
            return output

    def __arc_get_config_file_arg__(self):
        '''Helper function to return the config file argument'''
        if self.config['ArcConfigFile']:
            return "-z " + self.config['ArcConfigFile']

        return ""

    def arc_submit(self, jdlpath, ce, verbose):
        '''ARC CE direct job submission'''

        # use the CREAM UI check as it's the same
        if not self.__cream_ui_check__():
            return

        # No longer need to specify CE if available in client.conf
        # if not ce:
        #    logger.warning('No CREAM CE endpoint specified')
        #    return

        # write to a temporary XML file as otherwise can't submit in parallel
        tmpstr = '/tmp/' + randomString() + '.arcsub.xml'
        cmd = 'arcsub %s -S org.nordugrid.gridftpjob -j %s' % (
            self.__arc_get_config_file_arg__(), tmpstr)
        exec_bin = True

        if verbose:
            cmd += ' -d DEBUG '

        if ce:
            cmd = cmd + ' -c %s' % ce

        cmd = '%s %s < /dev/null' % (cmd, jdlpath)

        logger.debug('job submit command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['SubmissionTimeout'])

        if output:
            output = "%s" % output.strip()
        rc = self.shell.system('rm ' + tmpstr)

        # Job submitted with jobid:
        # gsiftp://lcgce01.phy.bris.ac.uk:2811/jobs/vSoLDmvvEljnvnizHq7yZUKmABFKDmABFKDmCTGKDmABFKDmfN955m
        match = re.search(
            '(gsiftp:\/\/\S+:2811\/jobs\/[0-9A-Za-z_\.\-]+)$', output)

        # Job submitted with jobid: https://ce2.dur.scotgrid.ac.uk:8443/arex/..
        if not match:
            match = re.search(
                '(https:\/\/\S+:8443\/arex\/[0-9A-Za-z_\.\-]+)$', output)

        if match:
            logger.debug('job id: %s' % match.group(1))
            return match.group(1)
        else:
            logger.warning('Job submission failed.')
            return

    def arc_status(self, jobids, cedict):
        '''ARC CE job status query'''

        if not self.__cream_ui_check__():
            return ([], [])

        if not jobids:
            return ([], [])

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        cmd = 'arcstat'
        exec_bin = True

        cmd = '%s %s -i %s -j %s' % (
            cmd, self.__arc_get_config_file_arg__(), idsfile, self.config["ArcJobListFile"])
        logger.debug('job status command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 1, 255],
                                        timeout=self.config['StatusPollingTimeout'])
        jobInfoDict = {}

        if rc != 0:
            logger.warning(
                'jobs not found in XML file: arcsync will be executed to update the job information')
            self.__arc_sync__(cedict)

        if rc == 0 and output:
            jobInfoDict = self.__arc_parse_job_status__(output)

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        return jobInfoDict

    def __arc_parse_job_status__(self, log):
        '''Parsing job status report from CREAM CE status query'''

        # Job: gsiftp://lcgce01.phy.bris.ac.uk:2811/jobs/FowMDmswEljnvnizHq7yZUKmABFKDmABFKDmxbGKDmABFKDmlw9pKo
        # State: Finished (FINISHED)
        # Exit Code: 0

        # Job: https://ce2.dur.scotgrid.ac.uk:8443/arex/jNxMDmXTj7jnVDJaVq17x81mABFKDmABFKDmhfRKDmjBFKDmLaCRVn
        # State: Finished (terminal:client-stageout-possible)
        # Exit Code: 0

        jobInfoDict = {}
        jid = None

        for ln in log.split('\n'):

            ln.strip()

            # do we have a failed retrieval?
            if ln.find("Job not found") != -1:
                logger.warning("Could not find info for job id '%s'" % jid)
                jid = None
            elif ln.find("Job:") != -1 and ln.find("gsiftp") != -1:
                # new job info block
                jid = ln[ln.find("gsiftp"):].strip()
                jobInfoDict[jid] = {}
            elif ln.find("Job:") != -1 and ln.find("https") != -1:
                # new job info block
                jid = ln[ln.find("https"):].strip()
                jobInfoDict[jid] = {}

            # get info
            if ln.find("State:") != -1:
                jobInfoDict[jid]['State'] = ln[
                    ln.find("State:") + len("State:"):].strip()

            if ln.find("Exit Code:") != -1:
                jobInfoDict[jid]['Exit Code'] = ln[
                    ln.find("Exit Code:") + len("Exit Code:"):].strip()

            if ln.find("Job Error:") != -1:
                jobInfoDict[jid]['Job Error'] = ln[
                    ln.find("Job Error:") + len("Job Error:"):].strip()

        return jobInfoDict

    def __arc_sync__(self, cedict):
        '''Collect jobs to jobs.xml'''

        if cedict[0]:
            cmd = 'arcsync %s -j %s -f -c %s' % (self.__arc_get_config_file_arg__(
            ), self.config["ArcJobListFile"], ' -c '.join(cedict))
        else:
            cmd = 'arcsync %s -j %s -f ' % (
                self.__arc_get_config_file_arg__(), self.config["ArcJobListFile"])

        if not self.active:
            logger.warning('LCG plugin is not active.')
            return False
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        logger.debug('sync ARC jobs list with: %s' % cmd)
        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=True), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['StatusPollingTimeout'])
        if rc != 0:
            logger.error('Unable to sync ARC jobs. Error: %s' % output)

    def arc_get_output(self, jid, directory):
        '''ARC CE job output retrieval'''

        if not self.__cream_ui_check__():
            return (False, None)

        # construct URI list from ID and output from arcls
        cmd = 'arcls %s %s' % (self.__arc_get_config_file_arg__(), jid)
        exec_bin = True
        logger.debug('arcls command: %s' % cmd)
        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=exec_bin), cmd),
                                        allowed_exit=[0, 255],
                                        timeout=self.config['SubmissionTimeout'])
        if rc:
            logger.error(
                "Could not find directory associated with ARC job ID '%s'" % jid)
            return False

        # URI is JID + filename
        gfiles = []
        for uri in output.split("\n"):
            if len(uri) == 0:
                continue
            uri = jid + "/" + uri
            gf = GridftpFileIndex()
            gf.id = uri
            gfiles.append(gf)

        cache = GridftpSandboxCache()
        cache.middleware = 'GLITE'
        cache.vo = self.config['VirtualOrganisation']
        cache.uploaded_files = gfiles
        return cache.download(files=map(lambda x: x.id, gfiles), dest_dir=directory)

    def arc_purgeMultiple(self, jobids):
        '''ARC CE job purging'''

        if not self.__cream_ui_check__():
            return False

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        cmd = 'arcclean'
        exec_bin = True

        cmd = '%s %s -i %s -j %s' % (
            cmd, self.__arc_get_config_file_arg__(), idsfile, self.config["ArcJobListFile"])

        logger.debug('job purge command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        logger.debug(output)

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        if rc == 0:
            return True
        else:
            return False

    def arc_cancel(self, jobid):
        '''Cancel a job'''

        cmd = 'arckill'
        exec_bin = True

        if not self.active:
            logger.warning('LCG plugin is not active.')
            return False
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        cmd = '%s %s %s -j %s' % (cmd, str(
            jobid)[1:-1], self.__arc_get_config_file_arg__(), self.config["ArcJobListFile"])

        logger.debug('job cancel command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        if rc == 0:
            # job cancelling succeeded, try to remove the glite command logfile
            # if it exists
            self.__clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return True
        else:
            logger.warning("Failed to cancel job %s.\n%s" % (jobid, output))
            self.__print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return False

    def arc_cancelMultiple(self, jobids):
        '''Cancel multiple jobs in one LCG job cancellation call'''

        # compose a temporary file with job ids in it
        if not jobids:
            return True

        cmd = 'arckill'
        exec_bin = True

        if not self.active:
            logger.warning('LCG plugin is not active.')
            return False
        if not self.credential.isValid('01:00'):
            logger.warning('GRID proxy lifetime shorter than 1 hour')
            return False

        idsfile = tempfile.mktemp('.jids')
        with open(idsfile, 'w') as ids_file:
            ids_file.write('\n'.join(jobids) + '\n')

        # compose the cancel command
        cmd = '%s %s -i %s -j %s' % (
            cmd, self.__arc_get_config_file_arg__(), idsfile, self.config["ArcJobListFile"])

        logger.debug('job cancel command: %s' % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (
            self.__get_cmd_prefix_hack__(binary=exec_bin), cmd), allowed_exit=[0, 255])

        # clean up tempfile
        if os.path.exists(idsfile):
            os.remove(idsfile)

        if rc == 0:
            # job cancelling succeeded, try to remove the glite command logfile
            # if it exists
            self.__clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return True
        else:
            logger.warning("Failed to cancel jobs.\n%s" % output)
            self.__print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
            return False

    def arc_info(self):
        '''Run the arcinfo command'''

        cmd = 'arcinfo %s > /dev/null' % self.__arc_get_config_file_arg__()
        logger.debug("Running arcinfo command '%s'" % cmd)

        rc, output, m = self.shell.cmd1('%s%s' % (self.__get_cmd_prefix_hack__(binary=True), cmd),
                                        allowed_exit=[0, 1, 255],
                                        timeout=self.config['StatusPollingTimeout'])
        return rc, output




class IMonitoringService(object):

    """ Interface of the monitoring service.
    Each method correponds to particular events which occur at ganga client or job wrapper (worker node).
    The event() method is retained for backwards compatibility and other methods by default use it.
    However it will be removed in the future.
    """

    def __init__(self, job_info, config_info=None):
        """Initialize the monitoring service.

        If the monitoring service is created in the Ganga client then job_info
        is the original job object, and config_info is the original config\
        object returned by getConfig().

        If the monitoring service is created on the worker node then job_info is
        the return value of getJobInfo(), and config_info is the return value of
        getConfig().getEffectiveOptions(), i.e. a dictionary.

        In order to support existing monitoring classes, which do not use
        config_info, if getConfig() returns None, the constructor is called with
        only the job_info argument.
        """
        self.job_info = job_info
        self.config_info = config_info

    def start(self, **opts):
        """Application is about to start on the worker node.
        Called by: job wrapper.
        """
        return self.event('start')

    def progress(self, **opts):
        """Application execution is in progress (called periodically, several times a second).
        Called by: job wrapper. """
        return self.event('progress')

    def stop(self, exitcode, **opts):
        """Application execution finished.
        Called by: job wrapper. """
        return self.event('end', {'exitcode': exitcode})

    def prepare(self, **opts):
        """Preparation of a job.
        Called by: ganga client. """
        return self.event('prepare', {'job': self.job_info})

    def submitting(self, **opts):
        """Just before the submission of a job.
        Called by: ganga client. """
        return self.event('submitting', {'job': self.job_info})

    def submit(self, **opts):
        """Submission of a job.
        Called by: ganga client. """
        return self.event('submit', {'job': self.job_info})

    def complete(self, **opts):
        """Completion of a job.
        Called by: ganga client. """
        return self.event('complete', {'job': self.job_info})

    def fail(self, **opts):
        """Failure of a job.
        Called by: ganga client. """
        return self.event('fail', {'job': self.job_info})

    def kill(self, **opts):
        """Killing of a job.
        Called by: ganga client. """
        return self.event('kill', {'job': self.job_info})

    def rollback(self, **opts):
        """Rollback of a job to new state (caused by error during submission).
        Called by: ganga client. """
        return self.event('rollback', {'job': self.job_info})

    def getJobInfo(self):
        """ Return a static info object which static information about the job
        at submission time. Called by: ganga client.

        The info object is passed to the contructor. Info
        object may only contain the standard python types (such as lists,
        dictionaries, int, strings).  """

        return None

    @staticmethod
    def getConfig():
        """Return the config object for this class.

        By default this method returns None. If it is overridden to
        return a Config object then the configuration is made available
        consistently in the Ganga client and on the worker node via the instance
        variable config_info. See __init__().

        For example, in your IMonitoringService implementation you could add::
            @staticmethod
            def getConfig():
                from Ganga.Utility import Config
                return Config.getConfig("MyMS")

        N.B. this is a static method.
        """
        return None





from __future__ import print_function
from __future__ import absolute_import
from Ganga.Core.exceptions import GangaException
from Ganga.Utility.logging import getLogger
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName

from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList, makeGangaListByRef

# config_scope is namespace used for evaluating simple objects (e.g. File)
from Ganga.Utility.Config import config_scope

from Ganga.Utility.Plugin import PluginManagerError, allPlugins

from Ganga.GPIDev.Base.Objects import GangaObject
from Ganga.GPIDev.Schema import Schema, Version
from Ganga.GPIDev.Lib.GangaList.GangaList import makeGangaList

from .GangaRepository import SchemaVersionError

import xml.sax.saxutils

logger = getLogger()

##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: VStreamer.py,v 1.1.2.2 2009-07-14 09:20:22 ebke Exp $
##########################################################################

# dump object (job) to file f (or stdout) while ignoring the attribute
# 'ignore_subs'


class XMLFileError(GangaException):

    def __init__(self, excpt, message):
        GangaException.__init__(self, excpt, message)
        self.message = message
        self.excpt = excpt

    def __str__(self):
        if self.excpt:
            err = '(%s:%s)' % (type(self.excpt), self.excpt)
        else:
            err = ''
        return "XMLFileError: %s %s" % (self.message, err)

def _raw_to_file(j, fobj=None, ignore_subs=[]):
    vstreamer = VStreamer(out=fobj, selection=ignore_subs)
    vstreamer.begin_root()
    stripProxy(j).accept(vstreamer)
    vstreamer.end_root()

def to_file(j, fobj=None, ignore_subs=[]):
    #used to debug write problems - rcurrie
    #_raw_to_file(j, fobj, ignore_subs)
    #return
    _ignore_subs = [ignore_subs] if not isinstance(ignore_subs, list) else ignore_subs
    try:
        _raw_to_file(j, fobj, _ignore_subs)
    except Exception as err:
        logger.error("XML to-file error for file:\n%s" % (err))
        raise XMLFileError(err, "to-file error")

# Faster, but experimental version of to_file without accept()
# def to_file(j,f=None,ignore_subs=''):
#    sl = ["<root>"]
#    sl.extend(fastXML(j,'   ',ignore_subs=ignore_subs))
#    sl.append("</root>\n")
#    f.write("".join(sl))

# load object (job) from file f
# if len(errors) > 0 the object was not loaded correctly.
# Typical exceptions are:
# * SchemaVersionError (incompatible schema version)
# * PluginManagerError (necessary plugin not loaded)
# * IOError (problem on file reading)
# * AssertionError (corruption: multiple objects in <root>...</root>
# * Exception (probably corrupted data problem)

def _raw_from_file(f):
    # logger.debug('----------------------------')
    ###logger.debug('Parsing file: %s',f.name)
    xml_content = f.read()
    obj, errors = Loader().parse(xml_content)
    return obj, errors

def from_file(f):
    #return _raw_from_file(f)
    try:
        return _raw_from_file(f)
    except Exception as err:
        logger.error("XML from-file error for file:\n%s" % err)
        raise XMLFileError(err, "from-file error")

##########################################################################
# utilities


def escape(s):
    # s.replace('"', '\\"').replace("'", "\\'"))
    return xml.sax.saxutils.escape(s)


def unescape(s):
    return xml.sax.saxutils.unescape(s)

# An experimental, fast way to print a tree of Ganga Objects to file
# Unused at the moment


def fastXML(obj, indent='', ignore_subs=''):
    if hasattr(obj, "__iter__") and not hasattr(obj, "iteritems"):
        sl = ["\n", indent, "<sequence>", "\n"]
        for so in obj:
            sl.append(indent)
            sl.extend(fastXML(so, indent + ' ', ignore_subs))
        sl.append(indent)
        sl.append("</sequence>")
        return sl
    elif hasattr(obj, '_data'):
        v = obj._schema.version
        sl = ['\n', indent, '<class name="%s" version="%i.%i" category="%s">\n' % (getName(obj), v.major, v.minor, obj._category)]
        for k, o in obj.getNodeData().iteritems():
            if k != ignore_subs:
                try:
                    if not obj._schema[k]._meta["transient"]:
                        sl.append(indent)
                        sl.append('<attribute name="%s">' % k)
                        sl.extend(fastXML(o, indent + '  ', ignore_subs))
                        sl.append('</attribute>\n')
                except KeyError as err:
                    logger.debug("KeyError: %s" % err)
                    pass
        sl.append(indent)
        sl.append('</class>')
        return sl
    else:
        return ["<value>", escape(repr(stripProxy(obj))), "</value>"]

##########################################################################
# A visitor to print the object tree into XML.
#


class VStreamer(object):
    # Arguments:
    # out: file-like output stream where to print, default sys.stdout
    # selection: string specifying the name of properties which should not be printed
    # e.g. 'subjobs' - will not print subjobs

    def __init__(self, out=None, selection=[]):
        self.level = 0
        self.selection = selection
        if out is not None:
            self.out = out
        else:
            import sys
            self.out = sys.stdout

    def begin_root(self):
        print('<root>', file=self.out)

    def end_root(self):
        print('</root>', file=self.out)

    def indent(self):
        return ' ' * (self.level - 1) * 3

    def nodeBegin(self, node):
        self.level += 1
        s = node._schema
        print(self.indent(), '<class name="%s" version="%d.%d" category="%s">' % (s.name, s.version.major, s.version.minor, s.category), file=self.out)

    def nodeEnd(self, node):
        print(self.indent(), '</class>', file=self.out)
        self.level -= 1
        return

    def print_value(self, x):
        print('\n', self.indent(), '<value>%s</value>' % escape(repr(stripProxy(x))), file=self.out)

    def showAttribute(self, node, name):
        return not node._schema.getItem(name)['transient'] and (self.level > 1 or name not in self.selection)

    def simpleAttribute(self, node, name, value, sequence):
        if self.showAttribute(node, name):
            self.level += 1
            print(self.indent(), end=' ', file=self.out)
            print('<attribute name="%s">' % name, end=' ', file=self.out)
            if sequence:
                self.level += 1
                print(file=self.out)
                print(self.indent(), '<sequence>', file=self.out)
                for v in value:
                    self.acceptOptional(v)
                print(self.indent(), '</sequence>', file=self.out)
                self.level -= 1
                print(self.indent(), '</attribute>', file=self.out)
            else:
                self.level += 1
                self.print_value(value)
                self.level -= 1
                print(self.indent(), end=' ', file=self.out)
                print('</attribute>', file=self.out)
            self.level -= 1

    def sharedAttribute(self, node, name, value, sequence):
        self.simpleAttribute(node, name, value, sequence)

    def acceptOptional(self, s):
        self.level += 1
        if s is None:
            print(self.indent(), '<value>None</value>', file=self.out)
        else:
            if isType(s, str):
                print(self.indent(), '<value>%s</value>' % escape(repr(s)), file=self.out)
            elif hasattr(stripProxy(s), 'accept'):
                stripProxy(s).accept(self)
            elif isType(s, (list, tuple, GangaList)):
                print(self.indent(), '<sequence>', file=self.out)
                for sub_s in s:
                    self.acceptOptional(sub_s)
                print(self.indent(), '</sequence>', file=self.out)
            else:
                self.print_value(stripProxy(s))
        self.level -= 1

    def componentAttribute(self, node, name, subnode, sequence):
        if self.showAttribute(node, name):
            self.level += 1
            print(self.indent(), '<attribute name="%s">' % name, file=self.out)
            if sequence:
                self.level += 1
                print(self.indent(), '<sequence>', file=self.out)
                for s in subnode:
                    self.acceptOptional(s)
                print(self.indent(), '</sequence>', file=self.out)
                self.level -= 1
            else:
                self.acceptOptional(subnode)
            print(self.indent(), '</attribute>', file=self.out)
            self.level -= 1


##########################################################################
# XML Parser.

# Empty Ganga Object


class EmptyGangaObject(GangaObject):

    """Empty Ganga Object. Is used to construct incomplete jobs"""
    _schema = Schema(Version(0, 0), {})
    _name = "EmptyGangaObject"
    _category = "internal"
    _hidden = 1

    def __init__(self):
        super(EmptyGangaObject, self).__init__()


class Loader(object):

    """ Job object tree loader.
    """

    def __init__(self):
        self.stack = None  # we construct object tree using this stack
        # ignore nested XML elements in case of data errors at a higher level
        self.ignore_count = 0
        self.errors = []  # list of exception objects in case of data errors
        # buffer for <value> elements (evaled as python expressions)
        self.value_construct = None
        # buffer for building sequences (FIXME: what about nested sequences?)
        self.sequence_start = []

    def parse(self, s):
        """ Parse and load object from string s using internal XML parser (expat).
        """
        import xml.parsers.expat

        # 3 handler functions
        def start_element(name, attrs):
            #logger.debug('Start element: name=%s attrs=%s', name, attrs) #FIXME: for 2.4 use CurrentColumnNumber and CurrentLineNumber
            # if higher level element had error, ignore the corresponding part
            # of the XML tree as we go down
            if self.ignore_count:
                self.ignore_count += 1
                return

            # initialize object stack
            if name == 'root':
                assert self.stack is None, "duplicated <root> element"
                self.stack = []
                return

            assert not self.stack is None, "missing <root> element"

            # load a class, make empty object and push it as the current object
            # on the stack
            if name == 'class':
                try:
                    cls = allPlugins.find(attrs['category'], attrs['name'])
                except PluginManagerError as e:
                    self.errors.append(e)
                    #self.errors.append('Unknown class: %(name)s'%attrs)
                    obj = EmptyGangaObject()
                    # ignore all elemenents until the corresponding ending
                    # element (</class>) is reached
                    self.ignore_count = 1
                else:
                    version = Version(*[int(v) for v in attrs['version'].split('.')])
                    if not cls._schema.version.isCompatible(version):
                        attrs['currversion'] = '%s.%s' % (cls._schema.version.major, cls._schema.version.minor)
                        self.errors.append(SchemaVersionError('Incompatible schema of %(name)s, repository is %(version)s currently in use is %(currversion)s' % attrs))
                        obj = EmptyGangaObject()
                        # ignore all elemenents until the corresponding ending
                        # element (</class>) is reached
                        self.ignore_count = 1
                    else:
                        # make a new ganga object
                        obj = cls()
                self.stack.append(obj)

            # push the attribute name on the stack
            if name == 'attribute':
                self.stack.append(attrs['name'])

            # start value_contruct mode and initialize the value buffer
            if name == 'value':
                self.value_construct = ''

            # save a marker where the sequence begins on the stack
            if name == 'sequence':
                self.sequence_start.append(len(self.stack))

        def end_element(name):
            #logger.debug('End element: name=%s', name)

            # if higher level element had error, ignore the corresponding part
            # of the XML tree as we go up
            if self.ignore_count:
                self.ignore_count -= 1
                return

            # when </attribute> is seen the current object, attribute name and
            # value should be on top of the stack
            if name == 'attribute':
                value = self.stack.pop()
                aname = self.stack.pop()
                obj = self.stack[-1]
                # update the object's attribute
                obj.setNodeAttribute(aname, value)
                #logger.info("Setting: %s = %s" % (aname, value))

            # when </value> is seen the value_construct buffer (CDATA) should
            # be a python expression (e.g. quoted string)
            if name == 'value':
                # unescape the special characters
                s = unescape(self.value_construct)
                #logger.debug('string value: %s',s)
                val = eval(s, config_scope)
                #logger.debug('evaled value: %s type=%s',repr(val),type(val))
                self.stack.append(val)
                self.value_construct = None

            # when </sequence> is seen we remove last items from stack (as indicated by sequence_start)
            # we make a GangaList from these items and put it on stack
            if name == 'sequence':
                pos = self.sequence_start.pop()
                alist = makeGangaList(self.stack[pos:])
                del self.stack[pos:]
                self.stack.append(alist)

            # when </class> is seen we finish initializing the new object
            # by setting remaining attributes to their default values
            # the object stays on the stack (will be removed by </attribute> or
            # is a root object)
            if name == 'class':
                obj = self.stack[-1]
                for attr, item in obj._schema.allItems():
                    if not attr in obj.getNodeData():
                        #logger.info("Opening: %s" % attr)
                        if item._meta["sequence"] == 1:
                            obj.setNodeAttribute(attr, makeGangaListByRef(obj._schema.getDefaultValue(attr)))
                            #setattr(obj, attr, makeGangaListByRef(obj._schema.getDefaultValue(attr)))
                        else:
                            obj.setNodeAttribute(attr, obj._schema.getDefaultValue(attr))
                            #setattr(obj, attr, obj._schema.getDefaultValue(attr))

                #print("Constructed: %s" % getName(obj))

        def char_data(data):
            # char_data may be called many times in one CDATA section so we need to build up
            # the full buffer for <value>CDATA</value> section incrementally
            if self.value_construct is not None:
                ###logger.debug('char_data: append=%s',data)
                # FIXME: decode data
                self.value_construct += data

        # start parsing using callbacks
        p = xml.parsers.expat.ParserCreate()

        p.StartElementHandler = start_element
        p.EndElementHandler = end_element
        p.CharacterDataHandler = char_data

        p.Parse(s)

        if len(self.stack) != 1:
            self.errors.append(AssertionError('multiple objects inside <root> element'))

        obj = self.stack[-1]

        #logger.info("obj.__dict__: %s" % obj.__dict__)

        # Raise Exception if object is incomplete
        for attr, item in obj._schema.allItems():
            if not attr in obj.getNodeData():
                raise AssertionError("incomplete XML file")
        return obj, self.errors




# This class (SessionLockManager) encapsulates fcntl-based locking that works on NFS, AFS and locally.
# You can use
# python SessionLock.py {1|2}
# to run locking tests (run several instances in the same directory, from
# different machines)

from __future__ import print_function

import functools
import threading
from collections import namedtuple

import os
import time
import errno
import fcntl
import random
import datetime
import getpass

try:
    import cPickle as pickle
except ImportError:
    import pickle

from Ganga.Utility.logging import getLogger

from Ganga.Utility.Config.Config import getConfig, ConfigError
from Ganga.GPIDev.Base.Proxy import getName

try:
    from Ganga.Core.GangaThread import GangaThread
    from Ganga.Core.GangaRepository import RepositoryError
except ImportError:
    from threading import Thread

    print("IMPORT ERROR SHOULD NOT OCCUR IN PRODUCTION CODE!!!!!!!!!!!!!!!!!!!!!!")
    

    class GangaThread(Thread):
    
        def __init__(self, name):
            self.name = name
            super(GangaThread, self).__init__()

        def should_stop(self):
            return False


    class RepositoryError(Exception):
        pass


logger = getLogger()


session_lock_last = 0
session_expiration_timeout = 0
try:
    session_expiration_timeout = getConfig('Configuration')['DiskIOTimeout']
except ConfigError, err:
    session_expiratrion_timeout = 5

session_lock_refresher = None

def getGlobalLockRef(session_name, sdir, gfn, _on_afs):
    global session_lock_refresher
    if session_lock_refresher is None:
        try:
            os.close(SessionLockManager.delay_init_open(gfn))
            registerGlobalSessionFile(gfn)
        except OSError as err:
            logger.debug("Startup Lock Refresher Exception: %s" % err)
            raise RepositoryError(None, "Error on session file '%s' creation: %s" % (gfn, err))
        session_lock_refresher = SessionLockRefresher(session_name, sdir, gfn, None, _on_afs)
        session_lock_refresher.start()
    return session_lock_refresher

sessionFiles = []
sessionFileHandlers = []

def registerGlobalSessionFile(thisSessionFile):
    global sessionFiles
    if thisSessionFile not in sessionFiles:
        sessionFiles.append(thisSessionFile)


def registerGlobalSessionFileHandler(thisFileHandler):
    global sessionFileHandlers
    if thisFileHandler not in sessionFileHandlers:
        sessionFileHandlers.append(thisFileHandler)


def removeGlobalSessionFiles():
    global sessionFiles
    for i in sessionFiles:
        if os.path.isfile(i):
            if not i.endswith('global_lock'):
                #logger.debug( "Removing: %s" + i )
                os.unlink(i)


def removeGlobalSessionFileHandlers():
    global sessionFileHandlers
    for i in sessionFileHandlers:
        try:
            fcntl.lockf(i, fcntl.LOCK_UN)
        except Exception as err:
            logger.debug("Failed to unlock or close sessionfilehandler")
            logger.debug("%s" % err)


def getGlobalSessionFiles():
    global sessionFiles
    return sessionFiles


class SessionLockRefresher(GangaThread):

    def __init__(self, session_name, sdir, fn, repo, afs):
        GangaThread.__init__(self, name='SessionLockRefresher', critical=True)
        self.session_name = session_name
        self.sdir = sdir
        self.fns = [fn]
        self.repos = [repo]
        self.afs = afs
        self.FileCheckTimes = {}

    # This function attempts to grab the ctime from a file which should exist
    # As we don't want this to fail outright it attempts to re-read every 1s
    # for 30sec before failing
    def delayread(self, filename):
        value = None
        i = 0
        while value is None:
            try:
                value = os.stat(filename).st_ctime
            except OSError as x:
                # print "fail"
                value = None
                i = i + 1
                time.sleep(0.1)
                if i >= 60:  # 3000:
                    raise x
                else:
                    continue

        return value

    def run(self):

        try:
            while not self.should_stop():

                self.updateNow()
                self.checkAndReap()

                sleeptime = 1  # +random.random()

                for i in range(int(sleeptime * 200)):
                    if not self.should_stop():
                        time.sleep(0.05 + random.random() * 0.05)
        finally:
            #logger.debug("Finishing Monitoring Loop")
            self.unregister()

    def checkAndReap(self):
        # TODO: Check for services active/inactive
        try:
            for index in range(len(self.fns)):

                now = self.updateLocks(index)

                # Clear expired session files if monitoring is active
                self.clearDeadLocks(now)

        except Exception as x:
            logger.warning( "Internal exception in session lock thread: %s %s" % (getName(x), x))

    def updateNow(self):
        try:
            for index in range(len(self.fns)):
                now = self.updateLocks(index)
        except Exception as x:
            logger.warning("Internal exception in Updating session lock thread: %s %s" % ( getName(x), x))

    def updateLocks(self, index):

        this_index_file = self.fns[index]
        if this_index_file in self.FileCheckTimes.keys():
            if abs(self.FileCheckTimes[this_index_file]-time.time()) >= 3:
                now = self._reallyUpdateLocks(index)
                self.FileCheckTimes[this_index_file] = now
        else:
            now = self._reallyUpdateLocks(index)
            self.FileCheckTimes[this_index_file] = now

        return self.FileCheckTimes[this_index_file]

    def _reallyUpdateLocks(self, index, failCount=0):
        this_index_file = self.fns[index]
        now = None
        try:
            oldnow = self.delayread(this_index_file)
            os.system('touch %s' % this_index_file)
            now = self.delayread(this_index_file) # os.stat(self.fn).st_ctime
        except OSError as x:
            if x.errno != errno.ENOENT:
                logger.debug("Session file timestamp could not be updated! Locks could be lost!")
                if now is None and failCount < 4:
                    try:
                        logger.debug("Attempting to lock file again, unknown error:\n'%s'" % x)
                        import time
                        time.sleep(0.5)
                        failcount=failCount+1
                        now = self._reallyUpdateLocks(index, failcount)
                    except Exception as err:
                        now = -999.
                        logger.debug("Received another type of exception, failing to update lockfile: %s" % this_index_file)
                else:
                    logger.warning("Failed to update lock file: %s 5 times." % this_index_file)
                    logger.warning("This could be due to a filesystem problem, or multiple versions of ganga trying to access the same file")
                    now = -999.
            else:
                if self.repos[index] != None:
                    raise RepositoryError(self.repos[index],
                    "[SessionFileUpdate] Run: Own session file not found! Possibly deleted by another ganga session.\n\
                    Possible reasons could be that this computer has a very high load, or that the system clocks on computers running Ganga are not synchronized.\n\
                    On computers with very high load and on network filesystems, try to avoid running concurrent ganga sessions for long.\n '%s' : %s" % (this_index_file, x))
                else:
                    from Ganga.Core import GangaException
                    raise GangaException("Error Opening global .session file for this session: %s" % this_index_file)
        return now

    def clearDeadLocks(self, now):
        try:
            # Make list of sessions that are "alive"
            ls_sdir = os.listdir(self.sdir)
            session_files = [f for f in ls_sdir if f.endswith(".session") and f.find(str(os.getpid())) == -1]
            #metadata_lock_files = [f for f in ls_sdir if f.endswith("metadata.locks") and not f.find( str(os.getpid()) ) ]
            #all_lock_files = [f for f in ls_sdir if f.endswith(".locks") and not f.find( str(os.getpid()) ) ]
            # lock_files = [ s for s in all_lock_files and not in
            # metadata_lock_files ]
            lock_files = [f for f in ls_sdir if f.endswith(".locks") and f.find(str(os.getpid())) == -1]

            ## Loop over locks which aren't belonging to this session!
            for sf in session_files:
                joined_path = os.path.join(self.sdir, sf)
                #print("join: %s" % joined_path)
                if joined_path in self.fns:
                    continue
                mtm = os.stat(joined_path).st_mtime
                # print "%s: session %s delta is %s seconds" % (time.time(),
                # sf, now - mtm)
                global session_expiration_timeout
                if abs(float(now) - float(mtm)) > session_expiration_timeout:
                    logger.warning("Removing session %s because of inactivity (no update since %s seconds)" % (sf, now - mtm))
                    os.unlink(joined_path)
                    session_files.remove(sf)
                # elif now - mtm  > session_expiration_timeout/2:
                #    logger.warning("%s: Session %s is inactive (no update since %s seconds, removal after %s seconds)" % (time.time(), sf, now - mtm, session_expiration_timeout))

            # remove all lock files that do not belong to sessions that are
            # alive
            for f in lock_files:
                # Determine the session file which controls this lock file
                asf = f.split(".session")[0] + ".session"
                if not asf in session_files:
                    #logger.debug("Removing dead file %s" % (f) )
                    #logger.debug("Found, %s session files" % (session_files) )
                    #logger.debug("self.fns[%s] = %s " % ( index, self.fns[index] ) )
                    #logger.debug( "%s, %s" % ( self.sdir, f ) )
                    os.unlink(os.path.join(self.sdir, f))
        except OSError as x:
            # nothing really important, another process deleted the session
            # before we did.
            logger.debug("Unimportant OSError in cleaning locks: %s" % x)

        return

    def numberRepos(self):
        assert(len(self.fns) == len(self.repos))
        return len(self.fns)

    def addRepo(self, fn, repo):
        #logger.debug("Adding Repo: %s" % repo )
        self.repos.append(repo)
        #logger.debug("Adding fn: %s" % fn )
        self.fns.append(fn)

    def removeRepo(self, fn, repo):
        #logger.debug("Removing fn: %s" % fn )
        self.fns.remove(fn)
        #logger.debug("Removing fn: %s" % fn )
        self.repos.remove(repo)

        assert(len(self.fns) == len(self.repos))


def synchronised(f):
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._lock:
            return f(self, *args, **kwargs)
    return decorated


class SessionLockManager(object):

    """ Class with thread that keeps a global lock file that synchronizes
    ID and counter access across Ganga sessions.
    DEVELOPER WARNING: On NFS, files that are not locked with lockf (NOT flock) will 
    NOT be synchronized across clients, even if a global lock file is used!
    Interface:
        * startup() starts the session, automatically called on init
        * shutdown() stops the thread, FREES ALL LOCKS
        * make_new_ids(n) returns n new (locked) ids
        * lock_ids(ids) returns the ids that were successfully locked
        * release_ids(ids) returns the ids that were successfully released (now: all)
    All access to an instance of this class MUST be synchronized!
    Should ONLY raise RepositoryError (if possibly-corrupting errors are found)
    """

    LastCountAccess = namedtuple('LastCountAccess', ['time', 'val'])

    def mkdir(self, dn):
        """Make sure the given directory exists"""
        try:
            os.makedirs(dn)
        except OSError as x:
            if x.errno != errno.EEXIST:
                raise RepositoryError(
                    self.repo, "OSError on directory create: %s" % x)

    def __init__(self, repo, root, name, minimum_count=0):

        self.repo = repo
        self.mkdir(root)
        realpath = os.path.realpath(root)
        # Use the hostname (os.uname()[1])  and the current time in ms to construct the session filename.
        # TODO: Perhaps put the username here?
        global session_lock_refresher
        if session_lock_refresher is None:
            t = datetime.datetime.now()
            this_date = t.strftime("%H.%M_%A_%d_%B_%Y")
            session_name = ".".join(
                [os.uname()[1], str(this_date), "PID", str(os.getpid()), "session"])
            #session_name = ".".join([os.uname()[1], str(int(time.time()*1000)), str(os.getpid()), "session"])
        else:
            session_name = session_lock_refresher.session_name

        self.sdir = os.path.join(realpath, "sessions")
        self.gfn = os.path.join(self.sdir, session_name)
        self.fn = os.path.join(self.sdir, session_name + "." + name + ".locks")
        self.cntfn = os.path.join(realpath, name, "cnt")

        self.afs = (realpath[:4] == "/afs")
        self.locked = set()
        self.count = minimum_count
        self.session_name = session_name
        self.name = name
        self.realpath = realpath
        #logger.debug( "Initializing SessionLockManager: " + self.fn )
        self._lock = threading.RLock()
        self.last_count_access = None

    @staticmethod
    def delay_init_open(filename):
        value = None
        i = 0
        while value is None:
            try:
                value = os.open(filename, os.O_EXCL | os.O_CREAT | os.O_WRONLY)
            except OSError as x:
                time.sleep(0.1)
                if i >= 60:
                    raise x
                i += 1
        return value

    @synchronised
    def startup(self):
        self.last_count_access = None

        # Ensure directories exist
        self.mkdir(os.path.join(self.realpath, "sessions"))
        self.mkdir(os.path.join(self.realpath, self.name))

        # setup global lock
        self.global_lock_setup()
        self.global_lock_acquire()
        try:
            # setup counter file if it does not exist, read it if it does
            if not os.path.exists(self.cntfn):
                try:
                    fd = self.delay_init_open(self.cntfn)
                    try:
                        os.write(fd, "0")
                    finally:
                        os.close(fd)
                    #registerGlobalSessionFile( self.cntfn )
                except OSError as x:
                    if x.errno != errno.EEXIST:
                        raise RepositoryError(
                            self.repo, "OSError on count file create: %s" % x)
            try:
                self.count = max(self.count, self.cnt_read())
            except ValueError, err:
                logger.debug("Startup ValueError Exception: %s" % err)
                logger.error("Corrupt count file '%s'! Trying to recover..." % (self.cntfn))
            except OSError as err:
                logger.debug("Startup OSError Exception: %s" % err)
                raise RepositoryError(self.repo, "OSError on count file '%s' access!" % (self.cntfn))
            self.cnt_write()
            # Setup session file
            try:
                fd = self.delay_init_open(self.fn)
                try:
                    os.write(fd, pickle.dumps(set()))
                finally:
                    os.close(fd)
                registerGlobalSessionFile(self.fn)
            except OSError as err:
                logger.debug("Startup Session Exception: %s" % err)
                raise RepositoryError(self.repo, "Error on session file '%s' creation: %s" % (self.fn, err))

            session_lock_refresher = getGlobalLockRef(self.session_name, self.sdir, self.gfn, self.afs)

            session_lock_refresher.addRepo(self.fn, self.repo)
            self.session_write()
        finally:
            self.global_lock_release()

    def updateNow(self):
        global session_lock_refresher
        session_lock_refresher.updateNow()

    @synchronised
    def shutdown(self):
        """Shutdown the thread and locking system (on ganga shutdown or repo error)"""
        #logger.debug( "Shutting Down SessionLockManager, self.fn = %s" % (self.fn) )
        # print "Shutting Down SessionLock"
        self.locked = set()
        try:
            global session_lock_refresher
            session_lock_refresher.stop()
            if session_lock_refresher is not None:
                # try:
                session_lock_refresher.removeRepo(self.fn, self.repo)
                # except:
                #    pass
                #session_lock_refresher.removeRepo( self.fn, self.repo )
                if session_lock_refresher.numberRepos() <= 1:
                    session_lock_refresher = None
            #logger.debug("Session file '%s' deleted " % (self.fn))
            if not self.afs:
                os.close(self.lockfd)
            os.unlink(self.fn)
            #os.unlink(self.lockfn)
            # os.unlink(self.gfn)
        except OSError as x:
            logger.debug("Session file '%s' or '%s' was deleted already or removal failed: %s" % (self.fn, self.gfn, x))

    @staticmethod
    def delayopen_global(lockfn):
        value = None
        i = 0
        while value is None:
            try:
                value = os.open(lockfn, os.O_RDWR)
            except OSError as x:
                 # print "fail"
                value = None
                i = i + 1
                time.sleep(0.1)
                if i >= 60:  # 3000:
                    raise x
                else:
                    continue
        return value

    # Global lock function
    @synchronised
    def global_lock_setup(self):

        if self.afs:
            self.lockfn = os.path.join(self.sdir, "global_lock")
            lock_path = str(self.lockfn) + '.afs'
            lock_file = os.path.join(lock_path, "lock_file")
            try:
                if not os.path.exists(lock_path):
                    os.makedirs(lock_path)
                if not os.path.isfile(lock_file):
                    with open(lock_file, "w"):
                        pass
            except Exception as err:
                logger.debug("Global Lock Setup Error: %s" % err)
        else:
            try:
                self.lockfn = os.path.join(self.sdir, "global_lock")
                if not os.path.isfile(self.lockfn):
                    with open(self.lockfn, "w"):
                        # create file (does not interfere with existing sessions)
                        pass
                self.lockfd = self.delayopen_global(self.lockfn)
                registerGlobalSessionFile(self.lockfn)
                registerGlobalSessionFileHandler(self.lockfd)
            except IOError as x:
                raise RepositoryError(self.repo, "Could not create lock file '%s': %s" % (self.lockfn, x))
            except OSError as x:
                raise RepositoryError(self.repo, "Could not open lock file '%s': %s" % (self.lockfn, x))

    @staticmethod
    def delay_lock_mod(lockfd, lock_mod):
        i = 0
        num_tries = 35
        while i <= num_tries:
            try:
                fcntl.lockf(lockfd, lock_mod)
            except IOError as x:
                time.sleep(0.1)
                i += 1
                if i == num_tries:  # If we're on the last try
                    raise x
                else:
                    continue
            else:
                break
        return

    def global_lock_acquire(self):
        try:
            if self.afs:

                lock_path = str(self.lockfn) + '.afs'
                lock_file = os.path.join(lock_path, "lock_file")

                def clean_path():
                    oldtime = os.stat(lock_file).st_ctime
                    nowtime = time.time()
                    if abs(int(nowtime) - oldtime) > 10:
                        #logger.debug( "cleaning global lock" )
                        os.system("fs setacl %s %s rlidwka" % (lock_path, getpass.getuser()))

                while True:
                    try:
                        if os.path.isfile(lock_file):
                            clean_path()
                        os.unlink(lock_file)
                        break
                    except Exception as err:
                        logger.debug("Global Lock aquire Exception: %s" % err)
                        time.sleep(0.01)

                os.system("fs setacl %s %s rliwka" % (lock_path, getpass.getuser()))

                while not os.path.isfile(lock_file):
                    with open(lock_file, "w"):
                        pass
                    time.sleep(0.01)

            else:
                self.delay_lock_mod(self.lockfd, fcntl.LOCK_EX)

            #logger.debug("global capture")
        except IOError as x:
            raise RepositoryError(self.repo, "IOError on lock ('%s'): %s" % (self.lockfn, x))

    def global_lock_release(self):
        try:
            if self.afs:
                lock_path = str(self.lockfn) + '.afs'
                os.system("fs setacl %s %s rlidwka" % (lock_path, getpass.getuser()))
            else:
                self.delay_lock_mod(self.lockfd, fcntl.LOCK_UN)

            #logger.debug("global release")
        except IOError as x:
            raise RepositoryError(self.repo, "IOError on unlock ('%s'): %s" % (self.lockfn, x))

    @synchronised
    def delay_session_open(self, filename):
        value = None
        i = 0
        while value is None:
            try:
                value = os.open(filename, os.O_RDONLY)
            except OSError as x:
                time.sleep(0.1)
                i = i + 1
                if i >= 60:
                    raise x
                else:
                    continue
        return value

    # Session read-write functions
    @synchronised
    def session_read(self, fn):
        """ Reads a session file and returns a set of IDs locked by that session.
            The global lock MUST be held for this function to work, although on NFS additional
            locking is done
            Raises RepositoryError if severe access problems occur (corruption otherwise!) """
        try:
            # This can fail (thats OK, file deleted in the meantime)
            fd = self.delay_session_open(fn)
            try:
                os.lseek(fd, 0, 0)
                if not self.afs:  # additional locking for NFS
                    fcntl.lockf(fd, fcntl.LOCK_SH)
                try:
                    # 00)) # read up to 1 MB (that is more than enough...)
                    return pickle.loads(os.read(fd, 1048576))
                except Exception as x:
                    logger.warning("corrupt or inaccessible session file '%s' - ignoring it (Exception %s %s)." % (fn, getName(x), x))
            finally:
                if not self.afs:  # additional locking for NFS
                    fcntl.lockf(fd, fcntl.LOCK_UN)
                os.close(fd)
        except OSError as x:
            if x.errno != errno.ENOENT:
                raise RepositoryError(
                    self.repo, "Error on session file access '%s': %s" % (fn, x))
        return set()

    # This function attempts to grab the ctime from a file which should exist
    # As we don't want this to fail outright it attempts to re-read every 1s
    # for 30sec before failing
    @synchronised
    def delayopen(self, filename):
        value = None
        i = 0
        while value is None:
            try:
                value = os.open(filename, os.O_WRONLY)
            except OSError as x:
                # print "fail"
                value = None
                i = i + 1
                time.sleep(0.1)
                if i >= 60:  # 3000:
                    raise x
                else:
                    continue

        return value

    # This function attempts to grab the ctime from a file which should exist
    # As we don't want this to fail outright it attempts to re-read every 1s
    # for 30sec before failing
    @synchronised
    def delaywrite(self, filename, data):
        value = None
        i = 0
        while value is None:
            try:
                value = os.write(filename, data)
            except OSError as x:
                # print "fail"
                value = None
                i = i + 1
                time.sleep(0.1)
                if i >= 60:  # 3000:
                    raise x
                else:
                    continue

        return value

    @synchronised
    def session_write(self):
        """ Writes the locked set to the session file. 
            The global lock MUST be held for this function to work, although on NFS additional
            locking is done
            Raises RepositoryError if session file is inaccessible """
        #logger.debug("Openining Session File: %s " % self.fn )
        try:
            # If this fails, we want to shutdown the repository (corruption
            # possible)
            fd = self.delayopen(self.fn)
            try:
                if not self.afs:
                    fcntl.lockf(fd, fcntl.LOCK_EX)
                self.delaywrite(fd, pickle.dumps(self.locked))
                if not self.afs:
                    fcntl.lockf(fd, fcntl.LOCK_UN)
                os.fsync(fd)
            finally:
                os.close(fd)
        except OSError as x:
            if x.errno != errno.ENOENT:
                raise RepositoryError(
                    self.repo, "Error on session file access '%s': %s" % (self.fn, x))
            else:
                #logger.debug( "File NOT found %s" %self.fn )
                raise RepositoryError(self.repo, "SessionWrite: Own session file not found! Possibly deleted by another ganga session.\n\
                                    Possible reasons could be that this computer has a very high load, or that the system clocks on computers running Ganga are not synchronized.\n\
                                    On computers with very high load and on network filesystems, try to avoid running concurrent ganga sessions for long.\n '%s' : %s" % (self.fn, x))
        except IOError as x:
            raise RepositoryError(
                self.repo, "Error on session file locking '%s': %s" % (self.fn, x))

    # counter read-write functions
    def cnt_read(self):
        """ Tries to read the counter file.
            Raises ValueError (invalid contents)
            Raises OSError (no access/does not exist)
            Raises RepositoryError (fatal)
            """
        try:
            if self.last_count_access is not None:
                last_count_time = self.last_count_access.time
                last_count_val = self.last_count_access.val
                last_time = os.stat(self.cntfn).st_ctime
                if last_time == last_count_time:
                    return last_count_val  # If the file hasn't changed since last check, return the cached value
            _output = None
            fd = os.open(self.cntfn, os.O_RDONLY)
            try:
                if not self.afs:  # additional locking for NFS
                    fcntl.lockf(fd, fcntl.LOCK_SH)
                # 100 bytes should be enough for any ID. Can raise ValueErrorr
                _output = int(os.read(fd, 100).split("\n")[0])
            finally:
                if not self.afs:  # additional locking for NFS
                    fcntl.lockf(fd, fcntl.LOCK_UN)
                os.close(fd)

                if _output != None:
                    self.last_count_access = SessionLockManager.LastCountAccess(os.stat(self.cntfn).st_ctime, _output)
                    return _output

        except OSError as x:
            if x.errno != errno.ENOENT:
                raise RepositoryError(
                    self.repo, "OSError on count file '%s' read: %s" % (self.cntfn, x))
            else:
                # This can be a recoverable error, depending on where it occurs
                raise
        except IOError as x:
            raise RepositoryError(
                self.repo, "Locking error on count file '%s' write: %s" % (self.cntfn, x))

    @synchronised
    def cnt_write(self):
        """ Writes the counter to the counter file. 
            The global lock MUST be held for this function to work correctly
            Raises OSError if count file is inaccessible """
        finished = False
        try:
            # If this fails, we want to shutdown the repository (corruption
            # possible)
            fd = os.open(self.cntfn, os.O_WRONLY)
            try:
                if not self.afs:
                    fcntl.lockf(fd, fcntl.LOCK_EX)
                os.write(fd, str(self.count) + "\n")
                if not self.afs:
                    fcntl.lockf(fd, fcntl.LOCK_UN)
            finally:
                os.close(fd)
            finished = True
        except OSError as x:
            if x.errno != errno.ENOENT:
                raise RepositoryError(self.repo, "OSError on count file '%s' write: %s" % (self.cntfn, x))
            else:
                raise RepositoryError(self.repo, "Count file '%s' not found! Repository was modified externally!" % (self.cntfn))
        except IOError as x:
            raise RepositoryError(self.repo, "Locking error on count file '%s' write: %s" % (self.cntfn, x))
        finally:
            if finished is True:
                self.last_count_access = SessionLockManager.LastCountAccess(os.stat(self.cntfn).st_ctime, self.count)

    # "User" functions
    @synchronised
    def make_new_ids(self, n):
        """ Locks the next n available ids and returns them as a list 
            Raise RepositoryError on fatal error"""
        self.global_lock_acquire()
        try:
            # Actualize count
            try:
                newcount = self.cnt_read()
            except ValueError:
                logger.warning("Corrupt job counter (possibly due to crash of another session)! Trying to recover...")
                newcount = self.count
            except OSError:
                raise RepositoryError(self.repo, "Job counter deleted! External modification to repository!")
            if not newcount >= self.count:
                #raise RepositoryError(self.repo, "Counter value decreased - logic error!")
                logger.warning("Internal counter increased - probably the count file was deleted.")
                newcount = self.count
            # someone used force_ids (for example old repository imports)
            if self.locked and max(self.locked) >= newcount:
                newcount = max(self.locked) + 1
            ids = range(newcount, newcount + n)
            self.locked.update(ids)
            self.count = newcount + n
            self.cnt_write()
            self.session_write()
            return list(ids)
        finally:
            self.global_lock_release()

    def safe_LockCheck(self):
        global session_lock_last
        this_time = time.time()
        if session_lock_last == 0:
            session_lock_last = this_time
        global session_expiration_timeout
        _diff = abs(session_lock_last - this_time)
        if _diff > session_expiration_timeout * 0.5 or _diff < 1:
            session_expiration_timeout = this_time
            global session_lock_refresher
            if session_lock_refresher is not None:
                session_lock_refresher.checkAndReap()
            else:
                # self.reap_locks()
                pass

    @synchronised
    def lock_ids(self, ids):

        self.safe_LockCheck()

        #logger.debug( "locking: %s" % ids)
        ids = set(ids)
        self.global_lock_acquire()
        try:
            try:
                sessions = [
                    sn for sn in os.listdir(self.sdir) if sn.endswith(self.name + ".locks")]
            except OSError as x:
                raise RepositoryError(
                    self.repo, "Could not list session directory '%s'!" % (self.sdir))

            slocked = set()
            for session in sessions:
                sf = os.path.join(self.sdir, session)
                if sf == self.fn:
                    continue
                slocked.update(self.session_read(sf))
            #logger.debug( "locked: %s" % slocked)
            ids.difference_update(slocked)
            self.locked.update(ids)
            #logger.debug( "stored_lock: %s" % self.locked)
            self.session_write()
            #logger.debug( "list: %s" % list(ids))
            return list(ids)
        finally:
            self.global_lock_release()

    @synchronised
    def release_ids(self, ids):
        #logger.debug( "releasing : %s" % ids)
        self.global_lock_acquire()
        try:
            self.locked.difference_update(ids)
            self.session_write()
            #logger.debug( "list: %s" % list(ids))
            return list(ids)
        finally:
            self.global_lock_release()

    @synchronised
    def check(self):
        self.global_lock_acquire()
        try:
            with open(self.cntfn) as f:
                newcount = int(f.readline())
            assert newcount >= self.count
            sessions = os.listdir(self.sdir)
            prevnames = set()
            for session in sessions:
                if not session.endswith(self.name + ".locks"):
                    continue
                try:
                    sf = os.path.join(self.sdir, session)
                    fd = -1
                    if not self.afs:
                        fd = os.open(sf, os.O_RDONLY)
                        fcntl.lockf(fd, fcntl.LOCK_SH)  # ONLY NFS
                    with open(sf) as sf_file:
                        names = pickle.load(sf_file)
                    if not self.afs and fd > 0:
                        fcntl.lockf(fd, fcntl.LOCK_UN)  # ONLY NFS
                        os.close(fd)
                except Exception as x:
                    logger.warning("CHECKER: session file %s corrupted: %s %s" % (session, getName(x), x))
                    continue
                if not len(names & prevnames) == 0:
                    logger.error("Double-locked stuff: " + names & prevnames)
                    assert False
                # prevnames.union_update(names) Should be alias to update but
                # not in some versions of python
                prevnames.update(names)

        finally:
            self.global_lock_release()

    @synchronised
    def get_lock_session(self, id):
        """get_lock_session(id)
        Tries to determine the session that holds the lock on id for information purposes, and return an informative string.
        Returns None on failure
        """
        self.safe_LockCheck()

        self.global_lock_acquire()
        try:
            sessions = [s for s in os.listdir(self.sdir) if s.endswith(self.name + ".locks")]
            for session in sessions:
                try:
                    sf = os.path.join(self.sdir, session)
                    fd = -1
                    if not self.afs:
                        fd = os.open(sf, os.O_RDONLY)
                        fcntl.lockf(fd, fcntl.LOCK_SH)  # ONLY NFS
                    with open(sf) as sf_file:
                        names = pickle.load(sf_file)
                    if not self.afs and fd > 0:
                        fcntl.lockf(fd, fcntl.LOCK_UN)  # ONLY NFS
                        os.close(fd)
                    if id in names:
                        return self.session_to_info(session)
                except Exception as err:
                    logger.debug("Get Lock Session Exception: %s" % err)
                    continue
        finally:
            self.global_lock_release()

    @synchronised
    def get_other_sessions(self):
        """get_session_list()
        Tries to determine the other sessions that are active and returns an informative string for each of them.
        """
        self.global_lock_acquire()
        try:
            sessions = [s for s in os.listdir(self.sdir) if s.endswith(
                ".session") and not os.path.join(self.sdir, s) == self.gfn]
            return [self.session_to_info(session) for session in sessions]
        finally:
            self.global_lock_release()

    @synchronised
    def reap_locks(self):
        """reap_locks() --> True/False
        Remotely clear all foreign locks from the session.
        WARNING: This is not nice.
        Returns True on success, False on error."""
        failed = False
        self.global_lock_acquire()
        try:
            #sessions = [s for s in os.listdir(self.sdir) if s.endswith(".session") and not os.path.join(self.sdir, s) == self.gfn]
            sessions = [s for s in os.listdir(self.sdir) if s.endswith(
                ".session") and int(str(s).split('.')[-2]) != int(os.getpid())]
            metadata_lock_files = [
                s for s in os.listdir(self.sdir) if s.endswith("metadata.locks")]
            all_lock_files = [
                s for s in os.listdir(self.sdir) if s.endswith(".locks")]
            lock_files = [
                s for s in all_lock_files if s not in metadata_lock_files]
            locks = [s for s in lock_files if int(
                str(s).split('.')[-4]) != int(os.getpid())]
            metadata_locks = [s for s in metadata_lock_files if int(
                str(s).split('.')[-5]) != int(os.getpid())]

            sessions.extend(locks)
            sessions.extend(metadata_locks)

#            for s in os.listdir(self.sdir):
#                if s.endswith(".session"):
#                    logger.debug( "PID: %s " % int(str(s).split('.')[-2]) )
#                    logger.debug( "OS PID: %s " % int(os.getpid()) )

            for session in sessions:
                try:
                    sf = os.path.join(self.sdir, session)
                    #logger.debug( "Modified %s" % os.stat(sf).st_ctime )
                    #logger.debug( "Time %s" % time.time() )
                    #logger.debug( "Diff_allowed %s" % session_expiration_timeout )
                    #logger.debug( "Diff %s" % (time.time() - os.stat(sf).st_ctime ) )
                    global session_expiration_timeout
                    if((time.time() - os.stat(sf).st_ctime) > session_expiration_timeout):
                        if(sf.endswith(".session")):
                            logger.debug("Reaping LockFile: %s" % (sf))
                        os.unlink(sf)
                except OSError as x:
                    failed = True
            return not failed
        finally:
            self.global_lock_release()

    def session_to_info(self, session):
        si = session.split(".")
        try:
            return "%s (pid %s) since %s" % (".".join(si[:-3]), si[-2], ".".join(si[-5:-3]))
        except Exception, err:
            logger.debug( "Session Info Exception: %s" % err)
            return session



'''Module that manages the database connection

This module is temporary, is planned to access the database 
via web service and not by direct access'''

import re

import psycopg2
import psycopg2.extras

from Ganga.Utility.Config import *
import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

# From psycogp2 documentation: In Python 2, if you want to uniformly receive 
# all your database input in Unicode, you can register the related typecasters 
# globally as soon as Psycopg is imported, and then forget about this story.
import psycopg2.extensions

psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)


def gridmon(query, params=None):
    '''Sends query to gridmon DB'''
    
    pattern = re.compile(r'\s+')
    query = re.sub(pattern, ' ', query.strip())
    
    # Connect to an existing database
    conn = psycopg2.connect(host = 'localhost',
                            database = 'gridmon',
                            user = getConfig('SuperB')['gridmon_user'],
                            password = getConfig('SuperB')['gridmon_pass'])
    
    # Open a cursor to perform database operations
    #cur = conn.cursor()
    cur = conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
    
    # log query
    logger.debug(cur.mogrify(query, params))
    
    # Query the database and obtain data as Python objects
    cur.execute("SET search_path = gridmonitor, pg_catalog;", )
    
    cur.execute(query, params)
    results = cur.fetchall()
    
    # Make the changes to the database persistent
    conn.commit()
    
    # Close communication with the database
    cur.close()
    conn.close()
    
    return results

def read(query, params=None):
    '''Sends SELECT query to sbk DB, returns all the results'''
    
    pattern = re.compile(r'\s+')
    query = re.sub(pattern, ' ', query.strip())
    
    # Connect to an existing database
    conn = psycopg2.connect(host = 'localhost',
                            database = 'sbk5',
                            user = getConfig('SuperB')['sbk_user'],
                            password = getConfig('SuperB')['sbk_pass'])
    
    # Open a cursor to perform database operations
    #cur = conn.cursor()
    cur = conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
    
    # Register adapter and typecaster for dict-hstore conversions.
    psycopg2.extras.register_hstore(conn, unicode=True)
    
    # Query the database and obtain data as Python objects
    cur.execute("SET search_path = sbk, pg_catalog;", )
    
    # log query
    logger.debug(cur.mogrify(query, params))
    
    cur.execute(query, params)
    results = cur.fetchall()
    
    # Make the changes to the database persistent
    conn.commit()
    
    # Close communication with the database
    cur.close()
    conn.close()
    
    return results

def write(query, params=None):
    '''Sends WRITE, UPDATE, DELETE query to sbk DB, returns nothing.'''
    
    pattern = re.compile(r'\s+')
    query = re.sub(pattern, ' ', query.strip())
    
    # Connect to an existing database
    conn = psycopg2.connect(host = 'localhost',
                            database = 'sbk5',
                            user = getConfig('SuperB')['sbk_user'],
                            password = getConfig('SuperB')['sbk_pass'])
    
    # Open a cursor to perform database operations
    #cur = conn.cursor()
    cur = conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
    
    # Register adapter and typecaster for dict-hstore conversions.
    psycopg2.extras.register_hstore(conn, unicode=True)
    
    # Query the database and obtain data as Python objects
    cur.execute("SET search_path = sbk, pg_catalog;", )
    
    # log query
    logger.debug(cur.mogrify(query, params))
    
    cur.execute(query, params)
    
    # Make the changes to the database persistent
    conn.commit()
    
    # Close communication with the database
    cur.close()
    conn.close()



"""
Provides Persistency and the base class for all Registries
Also, a list of all Registries is kept here
"""

from Ganga.Core.GangaRepository.GangaRepository import GangaRepository, RepositoryError, InaccessibleObjectError, SchemaVersionError
from Ganga.Core.GangaRepository.Registry import RegistryError, RegistryAccessError, RegistryKeyError, RegistryLockError, ObjectNotInRegistryError
from Ganga.Core.GangaRepository import GangaRepositoryXML

allRegistries = {}


def addRegistry(registry):
    """ Add a registry to the global dict"""
    allRegistries[registry.name] = registry

def getRegistries():
    """ Get all registries from the global dict"""
    return allRegistries.values()

def getRegistry(name):
    """ Get a specific Registry from the global dict"""
    return allRegistries[name]

def getRegistrySlice(name):
    """ Get The registry slice wrapping the registry of choice"""
    return allRegistries[name].getSlice()

def getRegistryProxy(name):
    """ Get the proxied registry slice wrapping the registry of choice"""
    return allRegistries[name].getProxy()





import os
## NB parseCommandLine first then import Dirac!!
from DIRAC.Core.Base.Script import parseCommandLine
parseCommandLine()
from DIRAC.Interfaces.API.Dirac import Dirac
dirac=Dirac()
errmsg=''
file_name='###UPLOAD_FILE###'
file_label=file_name

compressed = ###COMPRESSED###
if compressed:
    import gzip
    file_name += '.gz'
    f_in = open(file_label, 'rb')
    f_out = gzip.open(file_name, 'wb')
    f_out.writelines(f_in)
    f_out.close()
    f_in.close()

lfn= os.path.join('###LFN_BASE###', file_name)
wildcard='###WILDCARD###'
storage_elements=###SEs###

from contextlib import closing
with closing(open('###LOCATIONSFILE_NAME###','ab')) as locationsfile:
    for se in storage_elements:
        try:
            result = dirac.addFile(lfn, file_name, se)
        except Exception as x:
            import sys
            sys.stdout.write('Exception running dirac.addFile command: %s' % str(x))
            break
        if result.get('OK',False) and lfn in result.get('Value',{'Successful':{}})['Successful']:
            import datetime
            guid = dirac.getMetadata(lfn)['Value']['Successful'][lfn]['GUID']
            locationsfile.write("DiracFile:::%s&&%s->%s:::['%s']:::%s\\n" % (wildcard, file_label, lfn, se, guid))
            locationsfile.flush()
            locationsfile.close()
            break
        errmsg+="(%s, %s), " % (se, result)
    else:
        locationsfile.write("DiracFile:::%s&&%s->###FAILED###:::File '%s' could not be uploaded to any SE (%s):::NotAvailable\\n" % (wildcard, file_label, file_name, errmsg))
        import sys
        sys.stdout.write('Could not upload file %s' % file_name)




import os
import Ganga.Utility.logging
import Ganga.Utility.Config

from Ganga.Utility.Config.Config import _after_bootstrap
from Ganga.Utility.logging import getLogger
logger = getLogger()

if not _after_bootstrap:

    configGaudi = Ganga.Utility.Config.makeConfig(
    'GAUDI', 'Generic GAUDI based parameters')

    dscrpt = 'The command used to make a CMT application.'
    configGaudi.addOption('make_cmd', 'make', dscrpt)

    dscrpt = 'Levels below InstallArea/[<platform>]/python to decend when looking for .py files to include'
    configGaudi.addOption('pyFileCollectionDepth', 2, dscrpt)

def getEnvironment(config=None):
    import sys
    import os.path
    import PACKAGE

    PACKAGE.standardSetup()
    return


def loadPlugins(config=None):
    #import Lib.Backends
    #import Lib.Checkers
    pass
    #import Lib.Applications
    #import Lib.RTHandlers
    #import Lib.Datasets
    #import Lib.Datafiles
    #import Lib.Splitters



# Copyright information
__author__  = "Ole Weidner <oweidner@cct.lsu.edu>"
__date__    = "13 September 2009"
__version__ = "1.0"

COMPRESSED_TRANSFER = True

# Ganga GPIDev Imports
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import FileBuffer

# Compression lib
if COMPRESSED_TRANSFER:
    import tarfile

# Setup Utility Logger
import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

# Setup Utility Config
import Ganga.Utility.Config
config = Ganga.Utility.Config.makeConfig('SAGA','parameters of the SAGA backend')
config.addOption('remove_workdir', True, 'remove automatically the local working directory when the job completed')

# Other Utility Imports
import Ganga.Utility.util
import Ganga.Utility.logic

# SAGA Job Package Imports
import saga.job

# The threaded download manager
from Ganga.Core.GangaThread.MTRunner import MTRunner, Data, Algorithm
from Ganga.Core import GangaException

# Other classes that are part of the SAGA iBackend
from GangaSAGA.Lib.SAGA.SAGAFileTransferManager import SAGAFileTransferManager
from GangaSAGA.Lib.SAGA.SAGAWrapperScript import SAGAWrapperScript

# Python System Imports
import os,sys
import os.path,re,errno
import subprocess
import string

##############################################################################
## Helper functions for the multi-threaded file transfer manager
##
def start_saga_file_transfer_manager():
    global saga_file_transfer_manager
    saga_file_transfer_manager = None
    if not saga_file_transfer_manager:
        saga_file_transfer_manager = SAGAFileTransferManager(numThread=10)
        saga_file_transfer_manager.start()

def get_saga_file_transfer_manager():
    global saga_file_transfer_manager
    return saga_file_transfer_manager

# fire up the MTRunner
start_saga_file_transfer_manager()

##
##############################################################################


##############################################################################
##
class SAGA(IBackend):
    
    """Run jobs in the background using SAGA.
       The job is run localy or remotely - depending on which SAGA adaptor is selected. 
    """
   
    
    # Set category & unique backend name
    _category = 'backends'
    _name = 'SAGA'
    
    # Setup Job Attributes
    _schema = Schema(Version(1,2), {
        'status'      : SimpleItem(defvalue=None,typelist=None,protected=1,copyable=0,hidden=1,doc='*NOT USED*'),
        'actualCE'    : SimpleItem(defvalue='',protected=1,copyable=0,doc='Hostname where the job was submitted.'),
        'wrapper_pid' : SimpleItem(defvalue=-1,protected=1,copyable=0,hidden=1,doc='(internal) process id of the execution wrapper'),
                                    
        ## Public SAGA specific attributes - READ/WRITE - REQUIRED                        
        'jobservice_url' : SimpleItem(defvalue='fork://localhost',      doc='Resource manager URL that will be passed to SAGA'),
        'filesystem_url' : SimpleItem(defvalue='file://localhost/tmp/', doc='File System (working directory) URL that will be passed to SAGA'),
        
        ## Public SAGA specific attributes - READ/WRITE - OPTIONAL
        'workdir'             : SimpleItem(defvalue='',protected=0,copyable=0,doc='JSDL Attribute: Working Directory'),
        'queue'               : SimpleItem(defvalue='',protected=0,copyable=0,doc='JSDL Attribute: Queue.'),
        'number_of_processes' : SimpleItem(defvalue='1',protected=0,copyable=0,doc='JSDL Attribute: Number of Processes to Spawn'),
        'spmd_variation'      : SimpleItem(defvalue='single',protected=0,copyable=0,doc='JSDL Attribute: SPMD Variation.'),
        'allocation'          : SimpleItem(defvalue='',protected=0,copyable=0,doc='JSDL Attribute: Project / Allocation Name.'),
        'enable_compression'  : SimpleItem(defvalue=True,protected=0,copyable=0,doc='Creates an archive of the input sandbox before pre-staging.'),


        ## Hidden SAGA specific attributes - READ ONLY
        'saga_job_out' : SimpleItem(defvalue='',protected=1,copyable=0,doc='SAGA-internal URL to stdout.'),
        'saga_job_err' : SimpleItem(defvalue='',protected=1,copyable=0,doc='SAGA-internal URL to stderr.'),
        'saga_job_id'  : SimpleItem(defvalue='',protected=1,copyable=0,doc='SAGA-internal Job ID.'),
        
        ## Job-specific workdir. The full location of the work directory on the (remote) execution host
        ## will be filesystem_url + "/" + workdir_uuid + "/". Workdir_uuid can't be changed by the user.
        'workdir_uuid'  : SimpleItem(defvalue='',protected=1,copyable=0,doc='Job-specific working directory.'),
        
    })
    
    _GUIPrefs = [ { 'attribute' : 'jobservice_url', 'widget' : 'String' },
                  { 'attribute' : 'filesystem_url', 'widget' : 'String' },
                  { 'attribute' : 'id', 'widget' : 'Int' },
                  { 'attribute' : 'status' , 'widget' : 'String' },
                  { 'attribute' : 'exitcode', 'widget' : 'String' } ]
                  
    _GUIAdvancedPrefs = [ { 'attribute' : 'nice', 'widget' : 'String' },
                          { 'attribute' : 'exitcode', 'widget' : 'String' } ]    


    ##########################################################################
    ##
    def __init__(self):
      super(SAGA,self).__init__()


    ##########################################################################
    ## Tries to create the local working directories properly
    ##
    def setupworkdir(self, path):
        import shutil, errno
        
        job = self.getJobObject()
        # As a side effect, these functions create the
        # workspace directories if they don't exist
        input_wd_path = job.getInputWorkspace().getPath()
        output_wd_path = job.getOutputWorkspace().getPath()
        
        logger.debug('local workspace - input dir: %s', input_wd_path)
        logger.debug('local workspace - output dir: %s', output_wd_path)
        
        return True
        
        
    ##########################################################################
    ## Tries to pres-stage the input sandbox
    ##
    def prestagesandbox(self, jobconfig):
        
        job = self.getJobObject()
        mon = job.getMonitoringService()
        inw = job.getInputWorkspace() 
        
        sandbox_files = jobconfig.getSandboxFiles()
 
        logger.info("pre-staging files for saga job with id: %s", job.id)
             
        ## Compression is ENABLED. All input sandbox files are archived 
        ## into an uncompressed tarball which is then  transfered to the 
        ## execution host. This speeds up things a lot in many scenarios.
        ##
        if self.enable_compression:
            logger.info("  * adding %s user defined files to input sandbox", len(sandbox_files))
            
            import Ganga.Core.Sandbox as Sandbox
            from Ganga.GPIDev.Lib.File import File
            from Ganga.Core.Sandbox.WNSandbox import PYTHON_DIR
            import inspect

            fileutils = File( inspect.getsourcefile(Ganga.Utility.files), subdir=PYTHON_DIR )
            compressed_input_sandbox = job.createPackedInputSandbox(jobconfig.getSandboxFiles() + [ fileutils ] )

            try:
                for f in compressed_input_sandbox:
                    archive_url = saga.url("file://localhost/"+f)
                    logger.info("  * copying %s -> %s ", archive_url.url, self.filesystem_url+"/"+self.workdir_uuid+"/")
                    sf = saga.filesystem.file(archive_url.url)
                    sf.copy(self.filesystem_url+"/"+self.workdir_uuid+"/", saga.filesystem.Overwrite)
                
            except saga.exception as e:
                logger.error('exception caught while transfering file: %s', e.get_all_exceptions())
                job.updateStatus("failed")
                return False
               
        ## Compression is DISABLED. All input sandbox files are transfered 
        ## one-by-one to the execution host. This can be very slow, especially
        ## if SAGA is using the Globus GridFTP adaptor.
        else: 
            logger.info("prestaging %s input sandbox files", len(sandbox_files))
            for f in sandbox_files:
                try:
                    source = saga.url(f.name)
                    if(len(source.scheme) == 0 ):
                       source.scheme = "file"
                    if(len(source.host) == 0 ):
                        source.host = "localhost"

                    if(f.subdir != '.'): # create subdirectory & copy
                        target = saga.url(self.filesystem_url+"/"+self.workdir_uuid+"/"+f.subdir+"/")
                    
                        sd = saga.filesystem.directory(target, saga.filesystem.Create)
                        sf = saga.filesystem.file(source) 
                    
                        logger.info("  * copying %s -> %s ", source, target)
                        sf.copy(target, saga.filesystem.Overwrite)  # copy the file
                    
                    else: # copy to remote work dir
                        logger.info("  * copying %s -> %s ", source, self.filesystem_url+"/"+self.workdir_uuid+"/")
                        sf = saga.filesystem.file(source) 
                        sf.copy(self.filesystem_url+"/"+self.workdir_uuid+"/", saga.filesystem.Overwrite)
                    
                except saga.exception as e:
                    logger.error('exception caught while transfering file: %s', 
                    e.get_all_exceptions())
                    job.updateStatus("failed")
                    return False
            
        ##    
        ## copy the job script
        ##
        try:
            jobscript_path = saga.url("file://localhost/"+inw.getPath()+"/__jobscript__")
            if os.path.exists(jobscript_path.path):
                logger.info("  * copying %s -> %s ", jobscript_path.url, self.filesystem_url+"/"+self.workdir_uuid+"/")
                sf = saga.filesystem.file(jobscript_path.url)
                sf.copy(self.filesystem_url+"/"+self.workdir_uuid+"/", saga.filesystem.Overwrite)
        
        except saga.exception as e:
            logger.error('exception caught while transfering file: %s', e.get_all_exceptions())
            job.updateStatus("failed")
            return False
            
        return True
                
    ##########################################################################
    ## Tries to submit a ganga job through saga
    ##
    def submit(self, jobconfig, master_input_sandbox):
    
        # Make sure that all REQUIRED attributes are set
        if len(self.jobservice_url) == 0 :
            logger.error('jobservice_url attribute needs to be set')
            return 0
        if len(self.filesystem_url) == 0 :
            logger.error('filesystem_url attribute needs to be set')
            return 0
    
        self.setupworkdir(self.workdir)
                    
        try: 
            jd = saga.job.description()
            jd = self.makesagajobdesc(self.getJobObject(), jobconfig)
            
            if (True != self.prestagesandbox(jobconfig)):
                return False
                                    
            self.run(self.jobservice_url, jd)
                
        except saga.exception as e:
            logger.error('exception caught while submitting job: %s', e.get_all_exceptions())
            self.getJobObject().updateStatus("failed")
            return False
      
        return True # sets job to 'submitted'
        

    ##########################################################################
    ## Tries to resubmit an existing ganga job through saga
    ##
    def resubmit(self):                
        try: 
            jd = saga.job.description()
            jd = self.makesagajobdesc(self.getJobObject())
                  
            self.run(self.jobservice_url, jd)
                
        except saga.exception as e:
            logger.error('exception caught while re-submitting job: %s', e.get_all_exceptions())
            self.getJobObject().updateStatus("failed")
            return False
      
        return True # sets job to 'submitted'

           
      
    ##########################################################################
    ## Run a SAGA job
    ##
    def run(self, js_contact, jd):
        # actualCE is the same as the resource manager (for now)
        self.actualCE = self.jobservice_url;
        
        try:
            logger.info("submitting job to %s", self.jobservice_url);
            
            js_url = saga.url(self.jobservice_url)
            js =  saga.job.service(js_url)
            saga_job = js.create_job(jd)
      
            # execute the job
            saga_job.run()
            self.saga_job_id = saga_job.get_job_id()
            logger.info("job submitted with internal job id: %s", self.saga_job_id)
                  
        except saga.exception as e:
            logger.error('exception caught while submitting job: %s', e.get_all_messages())
            self.getJobObject().updateStatus("failed")
            return False
        
        return True

            
    ##########################################################################
    ## Tries to kill a running saga job
    ##        
    def kill(self):
        job = self.getJobObject()
        
        # create a new service object in order
        # to reconnect to the job we want to kill
        try :
            js_url = saga.url(job.backend.jobservice_url)
            js = saga.job.service(js_url)
            saga_job = js.get_job(job.backend.saga_job_id)
            
            # KILL KILL KILL
            saga_job.cancel()
                            
        except saga.exception as e:
            logger.error('exception caught while killing job: %s', 
                         e.get_all_messages())
            return False
            
        return True # sets job to 'killed'
        

    ##########################################################################
    ## Creates and returns a saga.job.description from a given
    ## job object
    ##
    def makesagajobdesc(self, job, jobconfig):

        ## We need a unique subdirectory per job to avoid input/output file clashes 
        ## The easiest way to do this is with a UUID-style directory name
        wd_uuid = "ganga-job-"
        #if job.name == '':
        #    wd_uuid += "noname-"
        #else:
        #    wd_uuid +=  job.name + "-"

        import uuid
        wd_uuid += str(uuid.uuid4()) 
        
        job.backend.workdir_uuid = wd_uuid
        
        ## Now we need to create a wrapper script on the fly. The wrapper
        ## script will be be transfered to the execution host and takes care
        ## of the archive unpacking as well as job monitoring / reporting.
        ws = SAGAWrapperScript()
                
        import inspect
        import Ganga.Core.Sandbox as Sandbox
        import Ganga.Utility as Utility
        ws.setInlineModules(inspect.getsource(Sandbox.WNSandbox))
        
        ws.setExecutable(jobconfig.getExeString())
        ws.setArguments(jobconfig.getArgStrings())
        ws.setOutputPatterns(jobconfig.outputbox)
        ws.setInputSandbox("_input_sandbox_"+str(job.id)+".tgz")
        
        text = ws.getScript()

        jd = saga.job.description()
        logger.debug("setting up new saga job with id: %s", job.id)
        
        # create jobscript in input sandbox
        from Ganga.GPIDev.Lib.File import FileBuffer
        jobscript = job.getInputWorkspace().writefile(FileBuffer('__jobscript__',text),executable=1)

        logger.debug("  * created new jobscript wrapper: %s", jobscript)

        # workdir
        if len(job.backend.workdir) != 0:        
            jd.working_directory = job.backend.workdir
            logger.debug("  * backend.workdir -> saga.workdir: %s", jd.working_directory)
        else: #default to the remote filesystem path component
            jd.working_directory = saga.url(self.filesystem_url+"/"+self.workdir_uuid+"/").path
            logger.debug("  * saga.workdir: %s (not given - extracted from 'filesystem_url')", jd.working_directory)
            
        # executable
        exe = jd.working_directory + '__jobscript__'
        jd.executable = exe #jobconfig.getExeString()
        logger.debug("  * application.exe -> saga.executable: %s", jd.executable)
        
        # arguments
        argList = jobconfig.getArgStrings()
        #for arg in job.application.args:
        #    argList.append( arg ) #"\\'%s\\'" % arg ) 
        if len(argList) != 0:
            jd.arguments = argList
            logger.debug("  * application.args -> saga.arguments: %s", jd.arguments)

        # environment
        envList = [] 
        for k, v in job.application.env.items():
            envList.append( k+"="+v ) #"\\'%s\\'" % arg ) 
        if len(envList) != 0:
            jd.environment = envList
            logger.debug("  * application.env -> saga.environment: %s", jd.environment)

        # queue
        if len(job.backend.queue) != 0:
            jd.queue = job.backend.queue
            logger.debug("  * backend.queue -> saga.queue: %s", jd.queue)

        # allocation 
        if len(job.backend.allocation) != 0:
            jd.job_project = [job.backend.allocation]
            logger.debug("  * backend.allocation -> saga.job_project: %s", jd.job_project)

        # spmd_variation
        if len(job.backend.spmd_variation) != 0:
            jd.spmd_variation = job.backend.spmd_variation
            logger.debug("  * backend.spmd_variation -> saga.spmd_variation: %s", jd.spmd_variation)

        # number_of_processes
        if len(job.backend.number_of_processes) != 0:
            jd.number_of_processes = job.backend.number_of_processes
            logger.debug("  * backend.number_of_processes -> saga.number_of_processes: %s", jd.number_of_processes)

        ## We have to create special filenames for stdout/stderr redirection
        ## To avoid name clashes, we append a UUID to the filename.               
        path_component = saga.url(self.filesystem_url+"/"+self.workdir_uuid+"/")

        try:
            d = saga.filesystem.directory(path_component, saga.filesystem.Create)
            logger.debug("  * created output/working directory on the remote system: %s", path_component)
        
        except saga.exception as e:
            logger.error('exception caught while creating output/working directory: %s', e.get_all_messages())
            self.getJobObject().updateStatus("failed")
            
        
        ## STDOUT
        self.saga_job_out = path_component.url + "/out.log"
        #jd.output =  saga.url(self.saga_job_out).path
        
        logger.debug("  * stdout should become available here: %s", 
            saga.url(self.saga_job_out).url)

        ## STDERR
        self.saga_job_err = path_component.url + "/err.log"
        #jd.error = saga.url(self.saga_job_err).path

        logger.debug("  * stderr should become available here: %s", 
            saga.url(self.saga_job_err).url)
        
        return jd
                

    ##########################################################################
    ## Method gets triggered by a ganga monitoring thread periodically 
    ## in order to update job information, like state, etc...
    ##
    @staticmethod
    def updateMonitoringInformation(jobs):
        
        for j in jobs:
            # Skip job status query in case the job is already in
            # 'completed' status. That should avoid a great amount of
            # overhead if there are many completed jobs in the list
            if j.status == 'completed':
                continue
                        
            # create a new service object in order
            # to reconnect to the job
            try :
                js_url = saga.url(j.backend.jobservice_url)

                # dirty hack no. 1: the default job (fork) adaptor
                # doesn't support re-connect! So if the URL starts with
                # something like fork://localhost... don't even try.
                if False: #((j.status != 'new') and (js_url.scheme == 'fork')):
                    # that's the only thing we can do right now. set the
                    # job status to completing and hope that it has actually
                    # completed and the output files are available
                    downloader = get_saga_file_transfer_manager()
                    downloader.addTask(j)
                    # download will set the status either to complete of failed
                   
                else:
                    js = saga.job.service(js_url)
                    saga_job = js.get_job(j.backend.saga_job_id)
            
                    # query saga job state
                    job_state = saga_job.get_state()
      
                    if job_state == saga.job.Done:
                        downloader = get_saga_file_transfer_manager()
                        downloader.addTask(j)
                        # download will set the status either to complete of failed

                    elif job_state == saga.job.Running:
                        if(j.status != 'running'):
                            j.updateStatus('running')
                        
                    elif job_state == saga.job.Failed:
                        if(j.status != 'failed'): 
                            j.updateStatus('failed')
                
            except saga.exception as e:
                logger.error('exception caught while updating job: %s', e.get_all_messages())
    
    ##########################################################################
    ## Make the monitoring function available to the update thread
    ##




    ##################################################
    ##
    def peek( self, filename = "", command = "" ):
      """
      Allow viewing of output files in job's work directory
      (i.e. while job is in 'running' state)
                                                                                
      Arguments other than self:
      filename : name of file to be viewed
                => Path specified relative to work directory
      command  : command to be used for file viewing
                                                                                
      Return value: None
      """
      #job = self.getJobObject()
      #topdir = self.workdir.rstrip( os.sep )
      #path = os.path.join( topdir, filename ).rstrip( os.sep )
      #job.viewFile( path = path, command = command )
      return None
      
#    def remove_workdir(self):
#        if config['remove_workdir']:
#            import shutil
#            try:
#                shutil.rmtree(self.workdir)
#            except OSError as x:
#                logger.warning('problem removing the workdir %s: %s',str(self.id),str(x))                        

      
##      
##############################################################################


##############################################################################
##
## Register a default runtime handler
from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from Ganga.Lib.Executable import RTHandler
allHandlers.add('Executable','SAGA', RTHandler)
 



from Ganga.Core.exceptions import SplitterError
from GangaGaudi.Lib.Splitters.GaudiInputDataSplitter import GaudiInputDataSplitter
from Ganga.GPIDev.Adapters.ISplitter import SplittingError
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.Utility.Config import getConfig
from Ganga.Utility.files import expandfilename
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName
import Ganga.Utility.logging
from Ganga.GPIDev.Lib.Job import Job
from GangaDirac.Lib.Files.DiracFile import DiracFile
import os
import copy
import pickle

from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Base.Filters import allComponentFilters
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList

logger = Ganga.Utility.logging.getLogger()


def getBackend():

    lhcbConfig = getConfig('LHCb')
    return lhcbConfig['SplitByFilesBackend']


class SplitByFiles(GaudiInputDataSplitter):

    """Splits a job into sub-jobs by partitioning the input data

    SplitByFiles can be used to split a job into multiple subjobs, where
    each subjob gets an unique subset of the inputdata files.
    """
    _name = 'SplitByFiles'
    _schema = Schema(Version(1, 0), {
        'filesPerJob': SimpleItem(defvalue=10,
                                  doc='Number of files per subjob',
                                  typelist=['int']),

        'maxFiles': SimpleItem(defvalue=None,
                               doc='Maximum number of files to use in a masterjob (None = all files)',
                               typelist=['int', 'type(None)']),

        'bulksubmit': SimpleItem(defvalue=False,
                                 doc='determines if subjobs are split '
                                 'server side in a "bulk" submission or '
                                 'split locally and submitted individually'),

        'ignoremissing': SimpleItem(defvalue=False,
                                    doc='Skip LFNs if they are not found '
                                    'in the LFC. This option is only used if'
                                    'jobs backend is Dirac'),

        'splitterBackend': SimpleItem(defvalue=getBackend(),
                                      doc='name of the backend algorithm to use for splitting',
                                      typelist=['str'], protected=1, visitable=0)
    })

    _exportmethods = ['split']

    def _attribute_filter__set__(self, name, value):
        if name is 'filesPerJob':
            if value > 100:
                logger.warning('filesPerJob exceeded DIRAC maximum')
                logger.warning('DIRAC has a maximum dataset limit of 100.')
                logger.warning(
                    'BE AWARE!... will set it to this maximum value at submit time if backend is Dirac')
        return value

    def _create_subjob(self, job, dataset):
        logger.debug("_create_subjob")
        datatmp = []

        logger.debug("dataset size: %s" % str(len(dataset)))
        #logger.debug( "dataset: %s" % str(dataset) )

        from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset

        if isinstance(dataset, LHCbDataset):
            for i in dataset:
                if isType(i, DiracFile):
                    datatmp.append(i)
                else:
                    logger.error("Unkown file-type %s, cannot perform split with file %s" % (type(i), str(i)))
                    from Ganga.Core.exceptions import GangaException
                    raise GangaException("Unkown file-type %s, cannot perform split with file %s" % (type(i), str(i)))
        elif type(dataset) == type([]) or isType(dataset, GangaList()):
            for this_file in dataset:
                if type(this_file) is str:
                    datatmp.append(allComponentFilters['gangafiles'](this_file, None))
                elif isType(this_file, IGangaFile):
                    datatmp.append(this_file)
                else:
                    logger.error("Unexpected type: %s" % str(type(this_file)))
                    logger.error("Wanted object to inherit from type: %s: %s" % (str(type(IGangaFile()))))
                    from Ganga.Core.exceptions import GangaException
                    x = GangaException("Unknown(unexpected) file object: %s" % this_file)
                    raise x
        elif type(dataset) is str:
            datatmp.append(DiracFile(lfn=dataset))
        else:
            logger.error("Unkown dataset type, cannot perform split here")
            from Ganga.Core.exceptions import GangaException
            logger.error("Dataset found: " + str(dataset))
            raise GangaException("Unkown dataset type, cannot perform split here")

        logger.debug("Creating new Job in Splitter")
        j = Job()
        logger.debug("Copying From Job")
        j.copyFrom(stripProxy(job), ['splitter', 'subjobs', 'inputdata', 'inputsandbox', 'inputfiles'])
        logger.debug("Unsetting Splitter")
        j.splitter = None
        #logger.debug("Unsetting Merger")
        #j.merger = None
        #j.inputsandbox = [] ## master added automatically
        #j.inputfiles = []
        logger.debug("Setting InputData")
        j.inputdata = LHCbDataset(files=datatmp[:],
                                  persistency=self.persistency,
                                  depth=self.depth)
        #j.inputdata.XMLCatalogueSlice = self.XMLCatalogueSlice
        logger.debug("Returning new subjob")
        return j

    # returns splitter generator
    def _splitter(self, job, inputdata):

        logger.debug("_splitter")

        indata = inputdata

        if indata is not None:

            self.depth = indata.depth
            self.persistency = indata.persistency
            self.XMLCatalogueSlice = indata.XMLCatalogueSlice

        else:

            self.depth = 0
            self.persistency = None
            self.XMLCatalogueSlice = None

        if stripProxy(job.backend).__module__.find('Dirac') > 0:

            logger.debug("found Dirac backend")

            if self.filesPerJob > 100:
                self.filesPerJob = 100  # see above warning
            logger.debug("indata: %s " % str(indata))

            if self.splitterBackend == "GangaDiracSplitter":
                from GangaDirac.Lib.Splitters.GangaSplitterUtils import GangaDiracSplitter
                outdata = GangaDiracSplitter(indata,
                                             self.filesPerJob,
                                             self.maxFiles,
                                             self.ignoremissing)
            elif self.splitterBackend == "OfflineGangaDiracSplitter":
                from GangaDirac.Lib.Splitters.OfflineGangaDiracSplitter import OfflineGangaDiracSplitter
                outdata = OfflineGangaDiracSplitter(indata,
                                                    self.filesPerJob,
                                                    self.maxFiles,
                                                    self.ignoremissing)
            elif self.splitterBackend == "splitInputDataBySize":
                from GangaLHCb.Lib.Splitters.LHCbSplitterUtils import DiracSizeSplitter
                outdata = DiracSizeSplitter(indata,
                                            self.filesPerJob,
                                            self.maxFiles,
                                            self.ignoremissing)
            elif self.splitterBackend == "splitInputData":
                indata = stripProxy(copy.deepcopy(inputdata))
                from GangaDirac.Lib.Splitters.SplitterUtils import DiracSplitter
                outdata = DiracSplitter(indata,
                                        self.filesPerJob,
                                        self.maxFiles,
                                        self.ignoremissing)
            else:
                raise SplitterError("Backend algorithm not selected!")

            logger.debug("outdata: %s " % str(outdata))
            return outdata
        else:
            logger.debug("Calling Parent Splitter as not on Dirac")
            return super(SplitByFiles, self)._splitter(job, indata)

    def split(self, job):
        logger.debug("split")
        if self.maxFiles == -1:
            self.maxFiles = None
        if self.bulksubmit == True:
            if stripProxy(job.backend).__module__.find('Dirac') > 0:
                logger.debug("Returning []")
                return []
        split_return = super(SplitByFiles, self).split(job)
        logger.debug("split_return: %s" % split_return)
        return split_return




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PACKAGE.py,v 1.1 2008-07-17 16:41:11 moscicki Exp $
##########################################################################

""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""

external_packages = {}

from Ganga.Utility.Setup import PackageSetup

setup = PackageSetup(external_packages)


def standardSetup(setup=setup):
    pass
